import IPython

IPY_MAJOR = IPython.version_info[0]
if IPY_MAJOR < 3:
    raise ImportError("IPython version %d is not supported." % IPY_MAJOR)

IPY3 = IPY_MAJOR == 3

# IPython underwent a major refactor between versions 3 and 4.  Many of the
# imports in version 4 have aliases to their old locations in 3, but they raise
# noisy deprecation warnings.  By conditionally importing here, we can support
# older versions without triggering warnings for users on new versions.
if IPY3:
    from IPython.nbformat import read
else:
    from nbformat import read


__all__ = ["read"]



import numpy as np
import pandas as pd
import unittest
import warnings

import empyrical as ep
from pyfolio.perf_attrib import (
    perf_attrib,
    create_perf_attrib_stats,
    _cumulative_returns_less_costs,
)


def _empyrical_compat_perf_attrib_result(index, columns, data):
    if ep.__version__ < "0.5.2":
        # Newer columns were added in empyrical v0.5.2. These exist in older
        # and newer empyrical:
        columns = [
            "risk_factor1",
            "risk_factor2",
            "common_returns",
            "specific_returns",
            "total_returns",
        ]
        data = {k: v for k, v in data.items() if k in columns}

    return pd.DataFrame(index=index, columns=columns, data=data)


def generate_toy_risk_model_output(
    start_date="2017-01-01", periods=10, num_styles=2
):
    """
    Generate toy risk model output.

    Parameters
    ----------
    start_date : str
        date to start generating toy data
    periods : int
        number of days for which to generate toy data

    Returns
    -------
    tuple of (returns, factor_returns, positions, factor_loadings)
    returns : pd.DataFrame
    factor_returns : pd.DataFrame
    """
    dts = pd.date_range(start_date, periods=periods)
    np.random.seed(123)
    tickers = ["AAPL", "TLT", "XOM"]
    styles = ["factor{}".format(i) for i in range(num_styles)]

    returns = pd.Series(index=dts, data=np.random.randn(periods)) / 100

    factor_returns = (
        pd.DataFrame(
            columns=styles,
            index=dts,
            data=np.random.randn(periods, len(styles)),
        )
        / 100
    )

    arrays = [dts, tickers]
    index = pd.MultiIndex.from_product(arrays, names=["dt", "ticker"])

    positions = pd.DataFrame(
        columns=tickers,
        index=dts,
        data=np.random.randint(100, size=(periods, len(tickers))),
    )
    positions["cash"] = np.zeros(periods)

    factor_loadings = pd.DataFrame(
        columns=styles,
        index=index,
        data=np.random.randn(periods * len(tickers), len(styles)),
    )

    return returns, positions, factor_returns, factor_loadings


def mock_transactions_from_positions(positions):
    # Compute the day-to-day diff of the positions frame, then collapse
    # that into a frame with one row per day per asset.
    transactions = (
        pd.melt(
            positions.diff().dropna().reset_index(),
            id_vars=["index"],
            var_name="symbol",
            value_name="amount",
        )
        .sort_values(["index", "symbol"])
        .set_index("index")
    )

    # Filter out positions that didn't actually change.
    transactions = transactions[transactions.amount != 0]

    # Tack on a price column.
    transactions["price"] = 100.0

    return transactions


class PerfAttribTestCase(unittest.TestCase):
    def test_perf_attrib_simple(self):

        start_date = "2017-01-01"
        periods = 2
        dts = pd.date_range(start_date, periods=periods)
        dts.name = "dt"

        tickers = ["stock1", "stock2"]
        styles = ["risk_factor1", "risk_factor2"]

        returns = pd.Series(data=[0.1, 0.1], index=dts)

        factor_returns = pd.DataFrame(
            columns=styles,
            index=dts,
            data={"risk_factor1": [0.1, 0.1], "risk_factor2": [0.1, 0.1]},
        )

        positions = pd.DataFrame(
            index=dts,
            data={"stock1": [20, 20], "stock2": [50, 50], "cash": [0, 0]},
        )

        index = pd.MultiIndex.from_product(
            [dts, tickers], names=["dt", "ticker"]
        )

        factor_loadings = pd.DataFrame(
            columns=styles,
            index=index,
            data={
                "risk_factor1": [0.25, 0.25, 0.25, 0.25],
                "risk_factor2": [0.25, 0.25, 0.25, 0.25],
            },
        )

        expected_perf_attrib_output = _empyrical_compat_perf_attrib_result(
            index=dts,
            columns=[
                "risk_factor1",
                "risk_factor2",
                "total_returns",
                "common_returns",
                "specific_returns",
                "tilt_returns",
                "timing_returns",
            ],
            data={
                "risk_factor1": [0.025, 0.025],
                "risk_factor2": [0.025, 0.025],
                "common_returns": [0.05, 0.05],
                "specific_returns": [0.05, 0.05],
                "tilt_returns": [0.05, 0.05],
                "timing_returns": [0.0, 0.0],
                "total_returns": returns,
            },
        )

        expected_exposures_portfolio = pd.DataFrame(
            index=dts,
            columns=["risk_factor1", "risk_factor2"],
            data={"risk_factor1": [0.25, 0.25], "risk_factor2": [0.25, 0.25]},
        )

        exposures_portfolio, perf_attrib_output = perf_attrib(
            returns, positions, factor_returns, factor_loadings
        )

        pd.util.testing.assert_frame_equal(
            expected_perf_attrib_output, perf_attrib_output
        )

        pd.util.testing.assert_frame_equal(
            expected_exposures_portfolio, exposures_portfolio
        )

        # test long and short positions
        positions = pd.DataFrame(
            index=dts,
            data={"stock1": [20, 20], "stock2": [-20, -20], "cash": [20, 20]},
        )

        exposures_portfolio, perf_attrib_output = perf_attrib(
            returns, positions, factor_returns, factor_loadings
        )

        expected_perf_attrib_output = _empyrical_compat_perf_attrib_result(
            index=dts,
            columns=[
                "risk_factor1",
                "risk_factor2",
                "total_returns",
                "common_returns",
                "specific_returns",
                "tilt_returns",
                "timing_returns",
            ],
            data={
                "risk_factor1": [0.0, 0.0],
                "risk_factor2": [0.0, 0.0],
                "common_returns": [0.0, 0.0],
                "specific_returns": [0.1, 0.1],
                "tilt_returns": [0.0, 0.0],
                "timing_returns": [0.0, 0.0],
                "total_returns": returns,
            },
        )

        expected_exposures_portfolio = pd.DataFrame(
            index=dts,
            columns=["risk_factor1", "risk_factor2"],
            data={"risk_factor1": [0.0, 0.0], "risk_factor2": [0.0, 0.0]},
        )

        pd.util.testing.assert_frame_equal(
            expected_perf_attrib_output, perf_attrib_output
        )

        pd.util.testing.assert_frame_equal(
            expected_exposures_portfolio, exposures_portfolio
        )

        perf_attrib_summary, exposures_summary = create_perf_attrib_stats(
            perf_attrib_output, exposures_portfolio
        )

        self.assertEqual(
            perf_attrib_summary["Annualized Specific Return"],
            perf_attrib_summary["Annualized Total Return"],
        )

        self.assertEqual(
            perf_attrib_summary["Cumulative Specific Return"],
            perf_attrib_summary["Total Returns"],
        )

        pd.util.testing.assert_frame_equal(
            exposures_summary,
            pd.DataFrame(
                0.0,
                index=["risk_factor1", "risk_factor2"],
                columns=[
                    "Average Risk Factor Exposure",
                    "Annualized Return",
                    "Cumulative Return",
                ],
            ),
        )

    def test_perf_attrib_regression(self):

        positions = pd.read_csv(
            "pyfolio/tests/test_data/positions.csv",
            index_col=0,
            parse_dates=True,
        )

        positions.columns = [
            int(col) if col != "cash" else col for col in positions.columns
        ]

        returns = pd.read_csv(
            "pyfolio/tests/test_data/returns.csv",
            index_col=0,
            parse_dates=True,
            header=None,
            squeeze=True,
        )

        factor_loadings = pd.read_csv(
            "pyfolio/tests/test_data/factor_loadings.csv",
            index_col=[0, 1],
            parse_dates=True,
        )

        factor_returns = pd.read_csv(
            "pyfolio/tests/test_data/factor_returns.csv",
            index_col=0,
            parse_dates=True,
        )

        residuals = pd.read_csv(
            "pyfolio/tests/test_data/residuals.csv",
            index_col=0,
            parse_dates=True,
        )

        residuals.columns = [int(col) for col in residuals.columns]

        intercepts = pd.read_csv(
            "pyfolio/tests/test_data/intercepts.csv",
            index_col=0,
            header=None,
            squeeze=True,
        )

        risk_exposures_portfolio, perf_attrib_output = perf_attrib(
            returns,
            positions,
            factor_returns,
            factor_loadings,
        )

        specific_returns = perf_attrib_output["specific_returns"]
        common_returns = perf_attrib_output["common_returns"]
        combined_returns = specific_returns + common_returns

        # since all returns are factor returns, common returns should be
        # equivalent to total returns, and specific returns should be 0
        pd.util.testing.assert_series_equal(
            returns, common_returns, check_names=False
        )

        self.assertTrue(np.isclose(specific_returns, 0).all())

        # specific and common returns combined should equal total returns
        pd.util.testing.assert_series_equal(
            returns, combined_returns, check_names=False
        )

        # check that residuals + intercepts = specific returns
        self.assertTrue(np.isclose((residuals + intercepts), 0).all())

        # check that exposure * factor returns = common returns
        expected_common_returns = risk_exposures_portfolio.multiply(
            factor_returns, axis="rows"
        ).sum(axis="columns")

        pd.util.testing.assert_series_equal(
            expected_common_returns, common_returns, check_names=False
        )

        # since factor loadings are ones, portfolio risk exposures
        # should be ones
        pd.util.testing.assert_frame_equal(
            risk_exposures_portfolio,
            pd.DataFrame(
                np.ones_like(risk_exposures_portfolio),
                index=risk_exposures_portfolio.index,
                columns=risk_exposures_portfolio.columns,
            ),
        )

        perf_attrib_summary, exposures_summary = create_perf_attrib_stats(
            perf_attrib_output, risk_exposures_portfolio
        )

        self.assertEqual(
            ep.annual_return(specific_returns),
            perf_attrib_summary["Annualized Specific Return"],
        )

        self.assertEqual(
            ep.annual_return(common_returns),
            perf_attrib_summary["Annualized Common Return"],
        )

        self.assertEqual(
            ep.annual_return(combined_returns),
            perf_attrib_summary["Annualized Total Return"],
        )

        self.assertEqual(
            ep.sharpe_ratio(specific_returns),
            perf_attrib_summary["Specific Sharpe Ratio"],
        )

        self.assertEqual(
            ep.cum_returns_final(specific_returns),
            perf_attrib_summary["Cumulative Specific Return"],
        )

        self.assertEqual(
            ep.cum_returns_final(common_returns),
            perf_attrib_summary["Cumulative Common Return"],
        )

        self.assertEqual(
            ep.cum_returns_final(combined_returns),
            perf_attrib_summary["Total Returns"],
        )

        avg_factor_exposure = risk_exposures_portfolio.mean().rename(
            "Average Risk Factor Exposure"
        )
        pd.util.testing.assert_series_equal(
            avg_factor_exposure,
            exposures_summary["Average Risk Factor Exposure"],
        )

        cumulative_returns_by_factor = pd.Series(
            [
                ep.cum_returns_final(perf_attrib_output[c])
                for c in risk_exposures_portfolio.columns
            ],
            name="Cumulative Return",
            index=risk_exposures_portfolio.columns,
        )

        pd.util.testing.assert_series_equal(
            cumulative_returns_by_factor,
            exposures_summary["Cumulative Return"],
        )

        annualized_returns_by_factor = pd.Series(
            [
                ep.annual_return(perf_attrib_output[c])
                for c in risk_exposures_portfolio.columns
            ],
            name="Annualized Return",
            index=risk_exposures_portfolio.columns,
        )

        pd.util.testing.assert_series_equal(
            annualized_returns_by_factor,
            exposures_summary["Annualized Return"],
        )

    def test_missing_stocks_and_dates(self):

        (
            returns,
            positions,
            factor_returns,
            factor_loadings,
        ) = generate_toy_risk_model_output()

        # factor loadings missing a stock should raise a warning
        factor_loadings_missing_stocks = factor_loadings.drop(
            "TLT", level="ticker"
        )

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", UserWarning)

            perf_attrib(
                returns,
                positions,
                factor_returns,
                factor_loadings_missing_stocks,
            )
            # avoids test failure due to DeprecationWarning for pandas>=1.0, <1.1
            w_ = [warn for warn in w if issubclass(warn.category, UserWarning)]
            self.assertEqual(len(w_), 1)
            self.assertIn(
                "The following assets were missing factor loadings: "
                "['TLT']",
                str(w_[-1].message),
            )
            self.assertIn(
                "Ratio of assets missing: 0.333", str(w_[-1].message)
            )

            # missing dates should raise a warning
            missing_dates = ["2017-01-01", "2017-01-05"]
            factor_loadings_missing_dates = factor_loadings.drop(missing_dates)

            exposures, perf_attrib_data = perf_attrib(
                returns,
                positions,
                factor_returns,
                factor_loadings_missing_dates,
            )
            w_ = [warn for warn in w if issubclass(warn.category, UserWarning)]
            self.assertEqual(len(w_), 2)
            self.assertIn(
                "Could not find factor loadings for "
                "{} dates".format(len(missing_dates)),
                str(w_[-1].message),
            )

            for date in missing_dates:
                self.assertNotIn(date, exposures.index)
                self.assertNotIn(date, perf_attrib_data.index)

            # perf attrib should work if factor_returns already missing dates
            exposures, perf_attrib_data = perf_attrib(
                returns,
                positions,
                factor_returns.drop(pd.DatetimeIndex(missing_dates)),
                factor_loadings_missing_dates,
            )
            w_ = [warn for warn in w if issubclass(warn.category, UserWarning)]
            self.assertEqual(len(w_), 3)
            self.assertIn(
                "Could not find factor loadings for "
                "{} dates".format(len(missing_dates)),
                str(w_[-1].message),
            )

            for date in missing_dates:
                self.assertNotIn(date, exposures.index)
                self.assertNotIn(date, perf_attrib_data.index)

            # test missing stocks and dates
            factor_loadings_missing_both = factor_loadings_missing_dates.drop(
                "TLT", level="ticker"
            )

            exposures, perf_attrib_data = perf_attrib(
                returns,
                positions,
                factor_returns,
                factor_loadings_missing_both,
            )
            w_ = [warn for warn in w if issubclass(warn.category, UserWarning)]
            self.assertEqual(len(w_), 5)
            self.assertIn(
                "The following assets were missing factor loadings: "
                "['TLT']",
                str(w_[-2].message),
            )
            self.assertIn(
                "Ratio of assets missing: 0.333", str(w_[-2].message)
            )

            self.assertIn(
                "Could not find factor loadings for "
                "{} dates".format(len(missing_dates)),
                str(w_[-1].message),
            )
            for date in missing_dates:
                self.assertNotIn(date, exposures.index)
                self.assertNotIn(date, perf_attrib_data.index)

            # raise exception if all stocks are filtered out
            empty_factor_loadings = factor_loadings.drop(
                ["AAPL", "TLT", "XOM"], level="ticker"
            )

            with self.assertRaisesRegexp(
                ValueError, "No factor loadings were available"
            ):

                exposures, perf_attrib_data = perf_attrib(
                    returns, positions, factor_returns, empty_factor_loadings
                )

    def test_high_turnover_warning(self):
        (
            returns,
            positions,
            factor_returns,
            factor_loadings,
        ) = generate_toy_risk_model_output()

        # Mock the positions data to turn over the whole portfolio from
        # one asset to another every day (cycling every 3 days).
        positions.iloc[::3, :] = [100.0, 0.0, 0.0, 0.0]
        positions.iloc[1::3, :] = [0.0, 100.0, 0.0, 0.0]
        positions.iloc[2::3, :] = [0.0, 0.0, 100.0, 0.0]

        transactions = mock_transactions_from_positions(positions)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", UserWarning)

            perf_attrib(
                returns,
                positions,
                factor_returns,
                factor_loadings,
                transactions=transactions,
            )
            # avoids test failure due to DeprecationWarning for pandas>=1.0, <1.1
            w = [warn for warn in w if issubclass(warn.category, UserWarning)]

        self.assertEqual(len([w]), 1)
        self.assertIn(
            "This algorithm has relatively high turnover of its positions.",
            str(w[-1].message),
        )

    def test_cumulative_returns_less_costs(self):

        returns = pd.Series(
            [0.1] * 3, index=pd.date_range("2017-01-01", periods=3)
        )
        cost = pd.Series([0.001] * len(returns), index=returns.index)

        expected_returns = pd.Series([0.1, 0.21, 0.331], index=returns.index)
        pd.util.testing.assert_series_equal(
            expected_returns, _cumulative_returns_less_costs(returns, None)
        )

        expected_returns = pd.Series(
            [0.099000, 0.207801, 0.327373], index=returns.index
        )
        pd.util.testing.assert_series_equal(
            expected_returns, _cumulative_returns_less_costs(returns, cost)
        )



from unittest import TestCase
from parameterized import parameterized
from datetime import datetime
import pandas as pd

from pandas.testing import assert_frame_equal, assert_series_equal

from pyfolio.capacity import (
    days_to_liquidate_positions,
    get_max_days_to_liquidate_by_ticker,
    get_low_liquidity_transactions,
    daily_txns_with_bar_data,
    apply_slippage_penalty,
)


class CapacityTestCase(TestCase):
    dates = pd.date_range(start="2015-01-01", freq="D", periods=3)

    positions = pd.DataFrame(
        [[1.0, 3.0, 0.0], [0.0, 1.0, 1.0], [3.0, 0.0, 1.0]],
        columns=["A", "B", "cash"],
        index=dates,
    )

    columns = ["sid", "amount", "price", "symbol"]
    transactions = pd.DataFrame(
        data=[[1, 100000, 10, "A"]] * len(dates), columns=columns, index=dates
    )

    volume = pd.DataFrame(
        [[1.0, 3.0], [2.0, 2.0], [3.0, 1.0]], columns=["A", "B"], index=dates
    )

    volume.index.name = "dt"
    volume *= 1000000
    volume["market_data"] = "volume"

    price = pd.DataFrame(
        [[1.0, 1.0]] * len(dates), columns=["A", "B"], index=dates
    )
    price.index.name = "dt"
    price["market_data"] = "price"

    market_data = (
        pd.concat([volume, price])
        .reset_index()
        .set_index(["dt", "market_data"])
    )

    def test_days_to_liquidate_positions(self):
        dtlp = days_to_liquidate_positions(
            self.positions,
            self.market_data,
            max_bar_consumption=1,
            capital_base=1e6,
            mean_volume_window=1,
        )

        expected = pd.DataFrame(
            [[0.0, 0.5 / 3], [0.75 / 2, 0.0]],
            columns=["A", "B"],
            index=self.dates[1:],
        )
        assert_frame_equal(dtlp, expected)

    def test_get_max_days_to_liquidate_by_ticker(self):
        mdtl = get_max_days_to_liquidate_by_ticker(
            self.positions,
            self.market_data,
            max_bar_consumption=1,
            capital_base=1e6,
            mean_volume_window=1,
        )

        expected = pd.DataFrame(
            [
                [datetime(2015, 1, 3), 0.75 / 2, 75.0],
                [datetime(2015, 1, 2), 0.5 / 3, 50.0],
            ],
            columns=["date", "days_to_liquidate", "pos_alloc_pct"],
            index=["A", "B"],
        )
        expected.index.name = "symbol"

        assert_frame_equal(mdtl, expected)

    @parameterized.expand(
        [
            (
                pd.DataFrame(
                    [
                        [datetime(2015, 1, 1), 100.0],
                        [datetime(2015, 1, 2), 100.0],
                    ],
                    columns=["date", "max_pct_bar_consumed"],
                    index=["A", "B"],
                ),
                None,
            ),
            (
                pd.DataFrame(
                    [[datetime(2015, 1, 3), 100 / 3]],
                    columns=["date", "max_pct_bar_consumed"],
                    index=["A"],
                ),
                1,
            ),
        ]
    )
    def test_get_low_liquidity_transactions(self, expected, last_n_days):
        txn_daily = pd.DataFrame(
            data=[
                [1, 1000000, 1, "A"],
                [2, 2000000, 1, "B"],
                [1, 1000000, 1, "A"],
            ],
            columns=["sid", "amount", "price", "symbol"],
            index=self.dates,
        )

        llt = get_low_liquidity_transactions(
            txn_daily, self.market_data, last_n_days=last_n_days
        )
        expected.index.name = "symbol"
        assert_frame_equal(llt, expected)

    def test_daily_txns_with_bar_data(self):
        daily_txn = daily_txns_with_bar_data(
            self.transactions, self.market_data
        )

        columns = ["symbol", "amount", "price", "volume"]
        expected = pd.DataFrame(
            data=[
                ["A", 100000, 1.0, 1000000.0],
                ["A", 100000, 1.0, 2000000.0],
                ["A", 100000, 1.0, 3000000.0],
            ],
            columns=columns,
            index=self.dates,
        )

        assert_frame_equal(daily_txn, expected)

    @parameterized.expand(
        [
            (1000000, 1, [0.9995, 0.9999375, 0.99998611]),
            (10000000, 1, [0.95, 0.99375, 0.998611]),
            (100000, 1, [0.999995, 0.999999375, 0.9999998611]),
            (1000000, 0.1, [0.99995, 0.99999375, 0.999998611]),
        ]
    )
    def test_apply_slippage_penalty(
        self, starting_base, impact, expected_adj_returns
    ):
        returns = pd.Series([1.0, 1.0, 1.0], index=self.dates)
        daily_txn = daily_txns_with_bar_data(
            self.transactions, self.market_data
        )

        adj_returns = apply_slippage_penalty(
            returns, daily_txn, starting_base, 1000000, impact=impact
        )
        expected_adj_returns = pd.Series(
            expected_adj_returns, index=self.dates
        )

        assert_series_equal(adj_returns, expected_adj_returns)



kwargs = {
    'plot_rolling_returns': {'title': '累積報酬率 (倍)', 'ylabel': '累積報酬率', 'xlabel': ''},
    'plot_returns': {'title': '報酬率 (倍)', 'ylabel': '報酬率', 'xlabel': ''},
    'plot_annual_returns': {'title': '年報酬率', 'ylabel': '報酬率', 'xlabel': '年度'},
    'plot_monthly_returns_dist': {'title': '月報酬率分布', 'ylabel': '月分數', 'xlabel': '報酬率'},
    'plot_monthly_returns_heatmap': {'title': '每月報酬率 (%)', 'ylabel': '年度', 'xlabel': '月分'},
    'plot_rolling_beta': {'title': '滾動Beta (6-month)', 'ylabel': 'Beta', 'xlabel': ''},
    'plot_rolling_volatility': {'title': '滾動波動度 ({}-month)', 'ylabel': '波動度', 'xlabel': ''},
    'plot_rolling_sharpe': {'title': '滾動夏普值 ({}-month)', 'ylabel': '夏普值', 'xlabel': ''},
    'plot_drawdown_periods': {'title': '前 {top} 大回撤期間', 'ylabel': '累積報酬率', 'xlabel': ''},
    'plot_drawdown_underwater': {'title': '回撤圖表', 'ylabel': '虧損幅度', 'xlabel': ''},
    'plot_gross_leverage': {'title': '總槓桿率', 'ylabel': '槓桿比率', 'xlabel': ''},
    'plot_exposures': {'title': '暴露度', 'ylabel': '暴露度', 'xlabel': ''},
    'show_and_plot_top_positions': {'title': '投資組合歷史前10大持股', 'ylabel': '持股比例', 'xlabel': ''},
    'plot_long_short_holdings': {'title': '買進/放空持股 (家數)', 'ylabel': '持股', 'xlabel': ''},
    'plot_holdings': {'title': '總持股檔數', 'ylabel': '持股檔數', 'xlabel': ''},
    'plot_sector_allocations': {'title': '產業配置', 'ylabel': '配置比例', 'xlabel': ''},
    'plot_turnover': {'title': '每日周轉率 (倍)', 'ylabel': '周轉率', 'xlabel': ''},
    'plot_return_quantiles': {'title': '報酬率分位數', 'ylabel': '', 'xlabel': ["日", "周", "月"]},
    'plot_slippage_sweep': {'title': '累積報酬率 (考慮額外每金額的滑價)', 'ylabel': '累積報酬率 (倍)', 'xlabel': ''},
    'plot_slippage_sensitivity': {'title': '平均年化報酬率 (考慮額外每金額的滑價)', 'ylabel': '平均年化報酬率', 'xlabel': '每金額滑價 (bps)'},
    'plot_daily_turnover_hist': {'title': '每日周轉率分布', 'ylabel': '次數', 'xlabel': '周轉率'},
    'plot_capacity_sweep': {'title': '模擬不同資金規模下，滑價對夏普值的影響', 'ylabel': '夏普值', 'xlabel': '資金規模 ($百萬)'},
    'plot_daily_volume': {'title': '每日交易量', 'ylabel': '交易股數 (股)', 'xlabel': ''},
    'plot_monthly_returns_timeseries': {'title': '每月報酬率', 'ylabel': '報酬率', 'xlabel': ''},
    'plot_perf_stats': {'title': '績效衡量指標分布', 'ylabel': '', 'xlabel': ''},
    'plot_max_median_position_concentration': {'title': '買進/放空 最大&中位數持股比例', 'ylabel': '持股比例', 'xlabel': ''},
    'plot_txn_time_hist':{'title':'交易時間分布','ylabel':'成交比例','xlabel':''},
    'plot_round_trip_lifetimes':{'title':'持股方向與期間分布','ylabel':'','xlabel':''},
    'plot_prob_profit_trade':{'title':'賣出持股/回補放空獲利的機率分布','ylabel':'次數','xlabel':''}
}


#
# Copyright 2016 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pandas as pd


def map_transaction(txn):
    """
    Maps a single transaction row to a dictionary.

    Parameters
    ----------
    txn : pd.DataFrame
        A single transaction object to convert to a dictionary.

    Returns
    -------
    dict
        Mapped transaction.
    """

    if isinstance(txn["sid"], dict):
        sid = txn["sid"]["sid"]
        symbol = txn["sid"]["symbol"]
    else:
        sid = txn["sid"]
        symbol = txn["sid"]

    return {
        "sid": sid,
        "symbol": symbol,
        "price": txn["price"],
        "order_id": txn["order_id"],
        "amount": txn["amount"],
        "commission": txn["commission"],
        "dt": txn["dt"],
    }


def make_transaction_frame(transactions):
    """
    Formats a transaction DataFrame.

    Parameters
    ----------
    transactions : pd.DataFrame
        Contains improperly formatted transactional data.

    Returns
    -------
    df : pd.DataFrame
        Daily transaction volume and dollar ammount.
         - See full explanation in tears.create_full_tear_sheet.
    """

    transaction_list = []
    for dt in transactions.index:
        txns = transactions.loc[dt]
        if len(txns) == 0:
            continue

        for txn in txns:
            txn = map_transaction(txn)
            transaction_list.append(txn)
    df = pd.DataFrame(sorted(transaction_list, key=lambda x: x["dt"]))
    df["txn_dollars"] = -df["amount"] * df["price"]

    df.index = list(map(pd.Timestamp, df.dt.values))
    return df


def get_txn_vol(transactions):
    """
    Extract daily transaction data from set of transaction objects.

    Parameters
    ----------
    transactions : pd.DataFrame
        Time series containing one row per symbol (and potentially
        duplicate datetime indices) and columns for amount and
        price.

    Returns
    -------
    pd.DataFrame
        Daily transaction volume and number of shares.
         - See full explanation in tears.create_full_tear_sheet.
    """

    txn_norm = transactions.copy()
    txn_norm.index = txn_norm.index.normalize()
    amounts = txn_norm.amount.abs()
    prices = txn_norm.price
    values = amounts * prices
    daily_amounts = amounts.groupby(amounts.index).sum()
    daily_values = values.groupby(values.index).sum()
    daily_amounts.name = "txn_shares"
    daily_values.name = "txn_volume"
    return pd.concat([daily_values, daily_amounts], axis=1)


def adjust_returns_for_slippage(
    returns, positions, transactions, slippage_bps
):
    """
    Apply a slippage penalty for every dollar traded.

    Parameters
    ----------
    returns : pd.Series
        Daily returns of the strategy, noncumulative.
         - See full explanation in create_full_tear_sheet.
    positions : pd.DataFrame
        Daily net position values.
         - See full explanation in create_full_tear_sheet.
    transactions : pd.DataFrame
        Prices and amounts of executed trades. One row per trade.
         - See full explanation in create_full_tear_sheet.
    slippage_bps: int/float
        Basis points of slippage to apply.

    Returns
    -------
    pd.Series
        Time series of daily returns, adjusted for slippage.
    """

    slippage = 0.0001 * slippage_bps
    portfolio_value = positions.sum(axis=1)

    '''
    **修改pnl算法** 20230725
    pnl（profit and loss）應該要是前一天portfolio_value * 當天returns，
    原先程式碼是`pnl = portfolio_value * returns`，
    代表是當天的portfolio_value * 當天的returns。所以做以下修改：
    1.將當天portfolio_value/(1 + 當天的returns)得到前一天的portfolio_value。
    2.接著再把步驟1的結果乘上當天的returns就得到當天的pnl
    
    此外，當adjusted_pnl=pnl=0時，skew及Kurtosis算不出來，因nan_policy='propagate'，
    所以adjusted_returns先fillna(0)。
    
    註：
    1.這邊的pnl與run_algorithm()裡的pnl不同。這邊的pnl包含收取股利的部分，因portfolio_value
    包含股利（減資退還）的部分。然而run_algorithm()裡的pnl不包含股利（減資退還）。

    2.這樣修改至少達到原本預期的功能，但是還是不太嚴謹。如果多付了slippage，這樣其實每天的portfolio_value
    都會異動，只調整有交易日子的returns，不太對。待後續版本優化。

    3.所以最嚴謹方式還是再跑一次回測，這個僅供快速比較使用。
    '''
    #pnl = portfolio_value * returns                                     # 20230725 (by MRC) 
    pnl = portfolio_value * returns  / (1 + returns)                     # 20230725 (by MRC) 
    
    traded_value = get_txn_vol(transactions).txn_volume
    slippage_dollars = traded_value * slippage
    adjusted_pnl = pnl.add(-slippage_dollars, fill_value=0)
    adjusted_returns = returns * adjusted_pnl / pnl
    
    return adjusted_returns.fillna(0) # 20230725 (by MRC)


def get_turnover(positions, transactions, denominator="AGB"):
    """
     - Value of purchases and sales divided by either the actual
     gross book or the portfolio value for the time step.

    Parameters
    ----------
    positions : pd.DataFrame
        Contains daily position values including cash.
        - See full explanation in tears.create_full_tear_sheet
    transactions : pd.DataFrame
        Prices and amounts of executed trades. One row per trade.
        - See full explanation in tears.create_full_tear_sheet
    denominator : str, optional
        Either 'AGB' or 'portfolio_value', default AGB.
        - AGB (Actual gross book) is the gross market
        value (GMV) of the specific algo being analyzed.
        Swapping out an entire portfolio of stocks for
        another will yield 200% turnover, not 100%, since
        transactions are being made for both sides.
        - We use average of the previous and the current end-of-period
        AGB to avoid singularities when trading only into or
        out of an entire book in one trading period.
        - portfolio_value is the total value of the algo's
        positions end-of-period, including cash.

    Returns
    -------
    turnover_rate : pd.Series
        timeseries of portfolio turnover rates.
    """

    txn_vol = get_txn_vol(transactions)
    traded_value = txn_vol.txn_volume

    if denominator == "AGB":
        # Actual gross book is the same thing as the algo's GMV
        # We want our denom to be avg(AGB previous, AGB current)
        AGB = positions.drop("cash", axis=1).abs().sum(axis=1)
        denom = AGB.rolling(2).mean()

        # Since the first value of pd.rolling returns NaN, we
        # set our "day 0" AGB to 0.
        denom.iloc[0] = AGB.iloc[0] / 2
    elif denominator == "portfolio_value":
        denom = positions.sum(axis=1)
    else:
        raise ValueError(
            "Unexpected value for denominator '{}'. The "
            "denominator parameter must be either 'AGB'"
            " or 'portfolio_value'.".format(denominator)
        )

    denom.index = denom.index.normalize()
    turnover = traded_value.div(denom, axis="index")
    turnover = turnover.fillna(0)
    return turnover



from unittest import TestCase

import pandas as pd

from pandas.testing import assert_series_equal

from pyfolio.txn import get_turnover, adjust_returns_for_slippage


class TransactionsTestCase(TestCase):
    def test_get_turnover(self):
        """
        Tests turnover using a 20 day period.

        With no transactions, the turnover should be 0.

        with 200% of the AGB traded each day, the daily
        turnover rate should be 2.0.
        """
        dates = pd.date_range(start="2015-01-01", freq="D", periods=20)

        # In this test, there is one sid (0) and a cash column
        positions = pd.DataFrame(
            [[10.0, 10.0]] * len(dates), columns=[0, "cash"], index=dates
        )

        # Set every other non-cash position to 40
        positions[0][::2] = 40

        transactions = pd.DataFrame(
            data=[], columns=["sid", "amount", "price", "symbol"], index=dates
        )

        # Test with no transactions
        expected = pd.Series([0.0] * len(dates), index=dates)
        result = get_turnover(positions, transactions).asfreq("D")
        assert_series_equal(result, expected)

        transactions = pd.DataFrame(
            data=[[1, 1, 10, 0]] * len(dates) + [[2, -1, 10, 0]] * len(dates),
            columns=["sid", "amount", "price", "symbol"],
            index=dates.append(dates),
        ).sort_index()

        # Turnover is more on day 1, because the day 0 AGB is set to zero
        # in get_turnover. On most days, we get 0.8 because we have 20
        # transacted and mean(10, 40) = 25, so 20/25.
        expected = pd.Series([1.0] + [0.8] * (len(dates) - 1), index=dates)
        result = get_turnover(positions, transactions).asfreq("D")

        assert_series_equal(result, expected)

        # Test with denominator = 'portfolio_value'
        result = get_turnover(
            positions, transactions, denominator="portfolio_value"
        ).asfreq("D")

        # Our portfolio value alternates between $20 and $50 so turnover
        # should alternate between 20/20 = 1.0 and 20/50 = 0.4.
        expected = pd.Series(
            [0.4, 1.0] * (int((len(dates) - 1) / 2) + 1), index=dates
        )

        assert_series_equal(result, expected)

    def test_adjust_returns_for_slippage(self):
        dates = pd.date_range(start="2015-01-01", freq="D", periods=20)

        positions = pd.DataFrame(
            [[0.0, 10.0]] * len(dates), columns=[0, "cash"], index=dates
        )

        # 100% total, 50% average daily turnover
        transactions = pd.DataFrame(
            data=[[1, 1, 10, "A"]] * len(dates),
            columns=["sid", "amount", "price", "symbol"],
            index=dates,
        )

        returns = pd.Series([0.05] * len(dates), index=dates)
        # 0.001% slippage per dollar traded
        slippage_bps = 10
        expected = pd.Series([0.049] * len(dates), index=dates)

        result = adjust_returns_for_slippage(
            returns, positions, transactions, slippage_bps
        )

        assert_series_equal(result, expected)



from unittest import TestCase
from parameterized import parameterized
from collections import OrderedDict
import os
import gzip

from pandas import Series, DataFrame, date_range, Timestamp, read_csv
from pandas.testing import assert_frame_equal

from numpy import (
    arange,
    zeros_like,
    nan,
)

import warnings

from pyfolio.utils import (
    to_utc,
    to_series,
    check_intraday,
    detect_intraday,
    estimate_intraday,
)
from pyfolio.pos import (
    get_percent_alloc,
    extract_pos,
    get_sector_exposures,
    get_max_median_position_concentration,
)


class PositionsTestCase(TestCase):
    dates = date_range(start="2015-01-01", freq="D", periods=20)

    def test_get_percent_alloc(self):
        raw_data = arange(15, dtype=float).reshape(5, 3)
        # Make the first column negative to test absolute magnitudes.
        raw_data[:, 0] *= -1

        frame = DataFrame(
            raw_data,
            index=date_range("01-01-2015", freq="D", periods=5),
            columns=["A", "B", "C"],
        )

        result = get_percent_alloc(frame)
        expected_raw = zeros_like(raw_data)
        for idx, row in enumerate(raw_data):
            expected_raw[idx] = row / row.sum()

        expected = DataFrame(
            expected_raw,
            index=frame.index,
            columns=frame.columns,
        )

        assert_frame_equal(result, expected)

    def test_extract_pos(self):
        index_dup = [
            Timestamp("2015-06-08", tz="UTC"),
            Timestamp("2015-06-08", tz="UTC"),
            Timestamp("2015-06-09", tz="UTC"),
            Timestamp("2015-06-09", tz="UTC"),
        ]
        index = [
            Timestamp("2015-06-08", tz="UTC"),
            Timestamp("2015-06-09", tz="UTC"),
        ]

        positions = DataFrame(
            {
                "amount": [100.0, 200.0, 300.0, 400.0],
                "last_sale_price": [10.0, 20.0, 30.0, 40.0],
                "sid": [1, 2, 1, 2],
            },
            index=index_dup,
        )
        cash = Series([100.0, 200.0], index=index)

        result = extract_pos(positions, cash)

        expected = DataFrame(
            OrderedDict(
                [
                    (1, [100.0 * 10.0, 300.0 * 30.0]),
                    (2, [200.0 * 20.0, 400.0 * 40.0]),
                    ("cash", [100.0, 200.0]),
                ]
            ),
            index=index,
        )
        expected.index.name = "index"
        expected.columns.name = "sid"

        assert_frame_equal(result, expected)

    @parameterized.expand(
        [
            (
                DataFrame(
                    [[1.0, 2.0, 3.0, 10.0]] * len(dates),
                    columns=[0, 1, 2, "cash"],
                    index=dates,
                ),
                {0: "A", 1: "B", 2: "A"},
                DataFrame(
                    [[4.0, 2.0, 10.0]] * len(dates),
                    columns=["A", "B", "cash"],
                    index=dates,
                ),
                False,
            ),
            (
                DataFrame(
                    [[1.0, 2.0, 3.0, 10.0]] * len(dates),
                    columns=[0, 1, 2, "cash"],
                    index=dates,
                ),
                Series(index=[0, 1, 2], data=["A", "B", "A"]),
                DataFrame(
                    [[4.0, 2.0, 10.0]] * len(dates),
                    columns=["A", "B", "cash"],
                    index=dates,
                ),
                False,
            ),
            (
                DataFrame(
                    [[1.0, 2.0, 3.0, 10.0]] * len(dates),
                    columns=[0, 1, 2, "cash"],
                    index=dates,
                ),
                {0: "A", 1: "B"},
                DataFrame(
                    [[1.0, 2.0, 10.0]] * len(dates),
                    columns=["A", "B", "cash"],
                    index=dates,
                ),
                True,
            ),
        ]
    )
    def test_sector_exposure(
        self, positions, mapping, expected_sector_exposure, warning_expected
    ):
        """
        Tests sector exposure mapping and rollup.

        """
        with warnings.catch_warnings(record=True) as w:
            result_sector_exposure = get_sector_exposures(positions, mapping)

            assert_frame_equal(
                result_sector_exposure, expected_sector_exposure
            )
            # avoids test failure due to DeprecationWarning for pandas>=1.0, <1.1
            w_ = [warn for warn in w if issubclass(warn.category, UserWarning)]
            if warning_expected:
                self.assertEqual(len(w_), 1)
            else:
                self.assertEqual(len(w_), 0)

    @parameterized.expand(
        [
            (
                DataFrame(
                    [[1.0, 2.0, 3.0, 14.0]] * len(dates),
                    columns=[0, 1, 2, "cash"],
                    index=dates,
                ),
                DataFrame(
                    [[0.15, 0.1, nan, nan]] * len(dates),
                    columns=[
                        "max_long",
                        "median_long",
                        "median_short",
                        "max_short",
                    ],
                    index=dates,
                ),
            ),
            (
                DataFrame(
                    [[1.0, -2.0, -13.0, 15.0]] * len(dates),
                    columns=[0, 1, 2, "cash"],
                    index=dates,
                ),
                DataFrame(
                    [[1.0, 1.0, -7.5, -13.0]] * len(dates),
                    columns=[
                        "max_long",
                        "median_long",
                        "median_short",
                        "max_short",
                    ],
                    index=dates,
                ),
            ),
            (
                DataFrame(
                    [[nan, 2.0, nan, 8.0]] * len(dates),
                    columns=[0, 1, 2, "cash"],
                    index=dates,
                ),
                DataFrame(
                    [[0.2, 0.2, nan, nan]] * len(dates),
                    columns=[
                        "max_long",
                        "median_long",
                        "median_short",
                        "max_short",
                    ],
                    index=dates,
                ),
            ),
        ]
    )
    def test_max_median_exposure(self, positions, expected):
        alloc_summary = get_max_median_position_concentration(positions)
        assert_frame_equal(expected, alloc_summary)

    __location__ = os.path.realpath(
        os.path.join(os.getcwd(), os.path.dirname(__file__))
    )

    test_returns = read_csv(
        gzip.open(__location__ + "/test_data/test_returns.csv.gz"),
        index_col=0,
        parse_dates=True,
    )
    test_returns = to_series(to_utc(test_returns))
    test_txn = to_utc(
        read_csv(
            gzip.open(__location__ + "/test_data/test_txn.csv.gz"),
            index_col=0,
            parse_dates=True,
        )
    )
    test_pos = to_utc(
        read_csv(
            gzip.open(__location__ + "/test_data/test_pos.csv.gz"),
            index_col=0,
            parse_dates=True,
        )
    )

    @parameterized.expand(
        [
            (test_pos, test_txn, False),
            (test_pos.resample("1W").last(), test_txn, True),
        ]
    )
    def test_detect_intraday(self, positions, transactions, expected):
        detected = detect_intraday(positions, transactions, threshold=0.25)
        assert detected == expected

    @parameterized.expand(
        [
            ("infer", test_returns, test_pos, test_txn, test_pos),
            (False, test_returns, test_pos, test_txn, test_pos),
        ]
    )
    def test_check_intraday(
        self, estimate, returns, positions, transactions, expected
    ):
        detected = check_intraday(estimate, returns, positions, transactions)
        assert_frame_equal(detected, expected)

    @parameterized.expand(
        [
            (test_returns, test_pos, test_txn, (1506, 8)),
            (
                test_returns,
                test_pos.resample("1W").last(),
                test_txn,
                (1819, 8),
            ),
        ]
    )
    def test_estimate_intraday(
        self, returns, positions, transactions, expected
    ):
        intraday_pos = estimate_intraday(returns, positions, transactions)
        assert intraday_pos.shape == expected



#
# Copyright 2016 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from math import copysign
from re import split
import warnings
from collections import deque, OrderedDict
from zipline.master import get_prices
import matplotlib.pyplot as plt
plt.rcParams['font.sans-serif'] = ['Microsoft JhengHei'] 

import pandas as pd
import numpy as np
import os


from .utils import print_table, format_asset, extract_rets_pos_txn_from_zipline

PNL_STATS = [
    ("Total profit", lambda x: x.sum()),
    ("Gross profit", lambda x: x[x > 0].sum()),
    ("Gross loss", lambda x: x[x < 0].sum()),
    (
        "Profit factor",
        lambda x: x[x > 0].sum() / x[x < 0].abs().sum()
        if x[x < 0].abs().sum() != 0
        else np.nan,
    ),
    ("Avg. trade net profit", "mean"),
    ("Avg. winning trade", lambda x: x[x > 0].mean()),
    ("Avg. losing trade", lambda x: x[x < 0].mean()),
    (
        "Ratio Avg. Win:Avg. Loss",
        lambda x: x[x > 0].mean() / x[x < 0].abs().mean()
        if x[x < 0].abs().mean() != 0
        else np.nan,
    ),
    ("Largest winning trade", "max"),
    ("Largest losing trade", "min"),
]

SUMMARY_STATS = [
    ("Total number of round_trips", "count"),
    ("Percent profitable", lambda x: len(x[x > 0]) / float(len(x))),
    ("Winning round_trips", lambda x: len(x[x > 0])),
    ("Losing round_trips", lambda x: len(x[x < 0])),
    ("Even round_trips", lambda x: len(x[x == 0])),
]

RETURN_STATS = [
    ("Avg returns all round_trips", lambda x: x.mean()),
    ("Avg returns winning", lambda x: x[x > 0].mean()),
    ("Avg returns losing", lambda x: x[x < 0].mean()),
    ("Median returns all round_trips", lambda x: x.median()),
    ("Median returns winning", lambda x: x[x > 0].median()),
    ("Median returns losing", lambda x: x[x < 0].median()),
    ("Largest winning trade", "max"),
    ("Largest losing trade", "min"),
]

DURATION_STATS = [
    ("Avg duration", lambda x: x.mean()),
    ("Median duration", lambda x: x.median()),
    ("Longest duration", lambda x: x.max()),
    ("Shortest duration", lambda x: x.min())
    #  FIXME: Instead of x.max() - x.min() this should be
    #  rts.close_dt.max() - rts.open_dt.min() which is not
    #  available here. As it would require a new approach here
    #  that passes in multiple fields we disable these measures
    #  for now.
    #  ('Avg # round_trips per day', lambda x: float(len(x)) /
    #   (x.max() - x.min()).days),
    #  ('Avg # round_trips per month', lambda x: float(len(x)) /
    #   (((x.max() - x.min()).days) / APPROX_BDAYS_PER_MONTH)),
]


def agg_all_long_short(round_trips, col, stats_dict):
    stats_all = (
        round_trips.assign(ones=1)
        .groupby("ones")[col]
        .agg(stats_dict)
        .T.rename(columns={1.0: "All trades"})
    )
    stats_long_short = (
        round_trips.groupby("long")[col]
        .agg(stats_dict)
        .T.rename(columns={False: "Short trades", True: "Long trades"})
    )

    return stats_all.join(stats_long_short)


def _groupby_consecutive(txn, max_delta=pd.Timedelta("8h")):
    """Merge transactions of the same direction separated by less than
    max_delta time duration.

    Parameters
    ----------
    transactions : pd.DataFrame
        Prices and amounts of executed round_trips. One row per trade.
        - See full explanation in tears.create_full_tear_sheet

    max_delta : pandas.Timedelta (optional)
        Merge transactions in the same direction separated by less
        than max_delta time duration.


    Returns
    -------
    transactions : pd.DataFrame

    """
    def vwap(transaction):
        if transaction.amount.sum() == 0:
            warnings.warn("Zero transacted shares, setting vwap to nan.")
            return np.nan
        return (
            transaction.amount * transaction.price
        ).sum() / transaction.amount.sum()

    out = []
    for _, t in txn.groupby("symbol"):
        t = t.sort_index()
        t.index.name = "dt"
        t = t.reset_index()

        t["order_sign"] = t.amount > 0
        t["block_dir"] = (
            (t.order_sign.shift(1) != t.order_sign).astype(int).cumsum()
        )
        t["block_time"] = (
            ((t.dt.sub(t.dt.shift(1))) > max_delta).astype(int).cumsum()
        )
        grouped_price = t.groupby(["block_dir", "block_time"]).apply(vwap)
        grouped_price.name = "price"
        grouped_rest = t.groupby(["block_dir", "block_time"]).agg(
            {"amount": "sum", "symbol": "first", "dt": "first"}
        )

        grouped = grouped_rest.join(grouped_price)

        out.append(grouped)

    out = pd.concat(out)
    out = out.set_index("dt")
    return out

def extract_round_trips(transactions, portfolio_value:None):
    """Group transactions into "round trips". First, transactions are
    grouped by day and directionality. Then, long and short
    transactions are matched to create round-trip round_trips for which
    PnL, duration and returns are computed. Crossings where a position
    changes from long to short and vice-versa are handled correctly.

    Under the hood, we reconstruct the individual shares in a
    portfolio over time and match round_trips in a FIFO-order.

    For example, the following transactions would constitute one round trip:
    index                  amount   price    symbol
    2004-01-09 12:18:01    10       50      'AAPL'
    2004-01-09 15:12:53    10       100      'AAPL'
    2004-01-13 14:41:23    -10      100      'AAPL'
    2004-01-13 15:23:34    -10      200       'AAPL'

    First, the first two and last two round_trips will be merged into a two
    single transactions (computing the price via vwap). Then, during
    the portfolio reconstruction, the two resulting transactions will
    be merged and result in 1 round-trip trade with a PnL of
    (150 * 20) - (75 * 20) = 1500.

    Note, that round trips do not have to close out positions
    completely. For example, we could have removed the last
    transaction in the example above and still generated a round-trip
    over 10 shares with 10 shares left in the portfolio to be matched
    with a later transaction.

    Parameters
    ----------
    transactions : pd.DataFrame
        Prices and amounts of executed round_trips. One row per trade.
        - See full explanation in tears.create_full_tear_sheet

    portfolio_value : pd.Series (optional)
        Portfolio value (all net assets including cash) over time.
        Note that portfolio_value needs to beginning of day, so either
        use .shift() or positions.sum(axis='columns') / (1+returns).

    Returns
    -------
    round_trips : pd.DataFrame
        DataFrame with one row per round trip.  The returns column
        contains returns in respect to the portfolio value while
        rt_returns are the returns in regards to the invested capital
        into that partiulcar round-trip.
    """

    """"20230927 大幅修正roundtrips在遭遇股票股利,分割,現增等股數變動之情況 Terry
        20230928 微調在清倉當天才發現以前有split ratio的情況, 額外在清倉當天計算以前split ratio對持股總數的影響
        20231031 修正遭遇減資時的調整，此前版本無針對國內減資進行特別調整"""
    from zipline.data import bundles
    bundle = bundles.load('tquant')

    roundtrips = []

    for sym, trans_sym in transactions.groupby("symbol"):
        print(f"正在計算{trans_sym.symbol[0]}交易紀錄")

        trans_sym = trans_sym.sort_index()
        price_stack, dt_stack = deque(), deque()

        trans_sym["signed_price"] = trans_sym.price * np.sign(trans_sym.amount)
        trans_sym["abs_amount"] = trans_sym.amount.abs().astype(int)

        cal = bundle.equity_daily_bar_reader.sessions 
        initial_date = trans_sym.index[0]

        for dt, t in trans_sym.iterrows():

            split_ratio = 0
            period = cal[(cal >= str(initial_date.date())) & (cal <= str(dt.date()))]

            split_info = bundle.adjustment_reader.load_adjustments(
                                dates=period,
                                assets=pd.Int64Index([trans_sym.symbol[0].sid]),
                                should_include_splits=True,
                                should_include_mergers=True,
                                should_include_dividends=True,
                                adjustment_type='volume')['volume']

            if t.price < 0:
                warnings.warn(
                    "Negative price detected, ignoring for" "round-trip."
                )
                continue
            
            indiv_prices = [t.signed_price] * t.abs_amount

            if ((len(price_stack) == 0) or (copysign(1, price_stack[-1]) == copysign(1, t.amount))) and len(split_info) != 0:

                if len(split_info) == 1:

                    split_ratio = split_info[list(split_info.keys())[0]][0].value

                else: # 20230927若兩筆交易中間存有兩次以上的split ratio事件
                    all_ratio = [split_info[key][0].value for key in split_info.keys()]

                    from functools import reduce

                    split_ratio = reduce(lambda x, y: x * y, all_ratio)

                if split_ratio > 1: #20231030股票股利、股票分割 Terry

                    stock_dividends = int(len(price_stack) * (split_ratio - 1))

                    price_stack.extend([0] * stock_dividends)

                    price_stack.extend(indiv_prices)

                    dt_stack.extend([dt] * (len(indiv_prices) + stock_dividends))

                elif split_ratio < 1: #20231030減資 Terry

                    capital_reducion = pd.Series(price_stack).value_counts().apply(lambda x: int(x * split_ratio))

                    capital_dt = pd.Series(dt_stack).value_counts().apply(lambda x: int(x * split_ratio))

                    if int(len(price_stack)*split_ratio) != capital_reducion.sum():

                        adjust = capital_reducion.sum() - int(len(price_stack)*split_ratio)

                        capital_reducion[capital_reducion.idxmax()] -= adjust #減去經調整後買入成本數量最多的股票，對減資後影響更為平均

                        capital_dt[capital_dt.idxmax()] -= adjust

                    capital_reducion = capital_reducion.reindex(pd.Series(price_stack).unique()) #由於採用先進先出，因此將順序調成price_stack內的順序

                    capital_dt = capital_dt.reindex(pd.Series(dt_stack).unique())

                    price_stack = deque(float(key) for key, value in capital_reducion.items() for _ in range(value)) #展開成price_stack的形式

                    dt_stack = deque(key for key, value in capital_dt.items() for _ in range(value))

                    if len(price_stack) == 0: #20231111 避免個股在庫存為0時遇到減資調整，卻沒有新增股票進庫存 Terry
 
                        price_stack.extend(indiv_prices)
 
                        dt_stack.extend([dt] * (len(indiv_prices)))

            # 20230927若一開始沒倉位(price_stack=0)或是最新的成本價與建倉數正負號相同(意味加倉)，則將最新一期的成本價加入price_stack中
            elif ((len(price_stack) == 0) or (copysign(1, price_stack[-1]) == copysign(1, t.amount))) and len(split_info) == 0:
                
                price_stack.extend(indiv_prices)

                dt_stack.extend([dt] * len(indiv_prices))

            else:
                # 20230928若在出清當天之前有split ratio，則計算出賣出之前的split ratio對持股數的影響
                if (copysign(1, price_stack[-1]) != copysign(1, t.amount)) and len(split_info) != 0:

                    if len(split_info) == 1:

                        split_ratio = split_info[list(split_info.keys())[0]][0].value

                    else: 
                        all_ratio = [split_info[key][0].value for key in split_info.keys()]

                        from functools import reduce

                        split_ratio = reduce(lambda x, y: x * y, all_ratio)

                    if split_ratio > 1: #20231030股票股利、股票分割 Terry

                        stock_dividends = int(len(price_stack) * (split_ratio - 1))

                        price_stack.extend([0] * stock_dividends)

                        dt_stack.extend([dt] * stock_dividends)

                    elif split_ratio < 1: #20231030減資 Terry

                        capital_reducion = pd.Series(price_stack).value_counts().apply(lambda x: int(x * split_ratio))

                        capital_dt = pd.Series(dt_stack).value_counts().apply(lambda x: int(x * split_ratio))

                        if int(len(price_stack)*split_ratio) != capital_reducion.sum():

                            adjust = capital_reducion.sum() - int(len(price_stack)*split_ratio)

                            dt_adjust = capital_dt.sum() - int(len(dt_stack)*split_ratio)

                            capital_reducion[capital_reducion.idxmax()] -= adjust #減去經調整後買入成本數量最多的股票，對減資後影響更為平均

                            capital_dt[capital_dt.idxmax()] -= dt_adjust

                        capital_reducion = capital_reducion.reindex(pd.Series(price_stack).unique()) #由於採用先進先出，因此將順序調成price_stack內的順序

                        capital_dt = capital_dt.reindex(pd.Series(dt_stack).unique())

                        price_stack = deque(float(key) for key, value in capital_reducion.items() for _ in range(value)) #展開成price_stack的形式

                        dt_stack = deque(key for key, value in capital_dt.items() for _ in range(value))

                pnl = 0
                invested = 0
                cur_open_dts = []

                for price in indiv_prices:

                    if len(price_stack) != 0 and (copysign(1, price_stack[-1]) != copysign(1, price)):

                        prev_price = price_stack.popleft()
                        prev_dt = dt_stack.popleft()

                        pnl += -(price + prev_price)

                        cur_open_dts.append(prev_dt)
                        invested += abs(prev_price)

                    else:

                        price_stack.append(price)
                        dt_stack.append(dt)
                

                roundtrips.append(
                    {
                        "pnl": pnl,
                        "open_dt": cur_open_dts[0],
                        "close_dt": dt,
                        "long": price < 0,
                        "rt_returns": pnl / invested,
                        "symbol": sym,
                    }
                )
            initial_date = dt
            
    roundtrips = pd.DataFrame(roundtrips)

    roundtrips["duration"] = roundtrips["close_dt"].sub(roundtrips["open_dt"])

    if portfolio_value is not None:
        # Need to normalize so that we can join
        pv = pd.DataFrame(portfolio_value, columns=["portfolio_value"]).assign(
            date=portfolio_value.index
        )

        roundtrips["date"] = roundtrips.close_dt.apply(
            lambda x: x.replace(hour=0, minute=0, second=0)
        )

        roundtrips = (
            roundtrips.set_index("date")
            .join(pv.set_index("date"), lsuffix="_")
            .reset_index()
        )

        #20230920 移除原本join後是另設一個tmp的變數，但原本roundtrips["returns"] = tmp.pnl/tmp.portfolio_value會不對齊，導致出現pnl為負但returns為正，反之亦然的現象 Terry
        roundtrips["returns"] = roundtrips.pnl / roundtrips.portfolio_value
        roundtrips = roundtrips.drop(["date", "portfolio_value"], axis=1)
    print('投組交易紀錄計算完畢')

    return roundtrips


def add_closing_transactions(positions, transactions):
    #20230919 原本的計算方式為剩餘投組價值除以剩餘股數，反推出結算股價，但此方法若面臨到股票分割和股票股利問題就會嚴重錯誤
    #因此改用投組總價值/結束當天(回測結束日期的前一天)之收盤價，得出應該清倉的股數
    """
    Appends transactions that close out all positions at the end of
    the timespan covered by positions data. Utilizes pricing information
    in the positions DataFrame to determine closing price.

    Parameters
    ----------
    positions : pd.DataFrame
        The positions that the strategy takes over time.
    transactions : pd.DataFrame
        Prices and amounts of executed round_trips. One row per trade.
        - See full explanation in tears.create_full_tear_sheet

    Returns
    -------
    closed_txns : pd.DataFrame
        Transactions with closing transactions appended.
    """
    from zipline.data import bundles
    bundle = bundles.load('tquant')

    closed_txns = transactions[["symbol", "amount", "price"]]

    pos_at_end = positions.drop("cash", axis=1).iloc[-1]
    open_pos = pos_at_end.replace(0, np.nan).dropna()
    # Add closing round_trips one second after the close to be sure
    # they don't conflict with other round_trips executed at that time.
    end_dt = open_pos.name + pd.Timedelta(hours=13, minutes=30)

    for sym, ending_val in open_pos.iteritems():
        txn_sym = transactions[transactions.symbol == sym]

        close_price = bundle.equity_daily_bar_reader.load_raw_arrays(columns=['close'],
                                                start_date=pd.Timestamp(open_pos.name.date(), tz='utc'),
                                                end_date=pd.Timestamp(open_pos.name.date(), tz='utc'),
                                                assets=[sym.sid])[0][0][0] 

        ending_amount = int(ending_val / close_price)
        closing_txn = OrderedDict(
            [
                ("amount", -ending_amount),
                ("price", close_price),
                ("symbol", sym),
            ]
        )

        closing_txn = pd.DataFrame(closing_txn, index=[end_dt])
        closed_txns = closed_txns.append(closing_txn)

    closed_txns = closed_txns[closed_txns.amount != 0]
    return closed_txns


def apply_sector_mappings_to_round_trips(round_trips, sector_mappings):
    """
    Translates round trip symbols to sectors.

    Parameters
    ----------
    round_trips : pd.DataFrame
        DataFrame with one row per round trip trade.
        - See full explanation in round_trips.extract_round_trips
    sector_mappings : dict or pd.Series, optional
        Security identifier to sector mapping.
        Security ids as keys, sectors as values.

    Returns
    -------
    sector_round_trips : pd.DataFrame
        Round trips with symbol names replaced by sector names.
    """

    sector_round_trips = round_trips.copy()
    sector_round_trips.symbol = sector_round_trips.symbol.apply(
        lambda x: sector_mappings.get(x, "No Sector Mapping")
    )
    sector_round_trips = sector_round_trips.dropna(axis=0)

    return sector_round_trips


def gen_round_trip_stats(round_trips):
    """Generate various round-trip statistics.

    Parameters
    ----------
    round_trips : pd.DataFrame
        DataFrame with one row per round trip trade.
        - See full explanation in round_trips.extract_round_trips

    Returns
    -------
    stats : dict
       A dictionary where each value is a pandas DataFrame containing
       various round-trip statistics.

    See also
    --------
    round_trips.print_round_trip_stats
    """

    stats = {}
    stats["pnl"] = agg_all_long_short(round_trips, "pnl", PNL_STATS)
    stats["summary"] = agg_all_long_short(round_trips, "pnl", SUMMARY_STATS)
    stats["duration"] = agg_all_long_short(
        round_trips, "duration", DURATION_STATS
    )
    stats["returns"] = agg_all_long_short(round_trips, "returns", RETURN_STATS)

    stats["symbols"] = (
        round_trips.groupby("symbol")["returns"].agg(RETURN_STATS).T
    )

    return stats


def print_round_trip_stats(round_trips, hide_pos=False):
    """Print various round-trip statistics. Tries to pretty-print tables
    with HTML output if run inside IPython NB.

    Parameters
    ----------
    round_trips : pd.DataFrame
        DataFrame with one row per round trip trade.
        - See full explanation in round_trips.extract_round_trips

    See also
    --------
    round_trips.gen_round_trip_stats
    """

    stats = gen_round_trip_stats(round_trips)

    print_table(
        stats["summary"], float_format="{:.2f}".format, name="Summary stats"
    )
    print_table(stats["pnl"], float_format="${:.2f}".format, name="PnL stats")
    print_table(
        stats["duration"], float_format="{:.2f}".format, name="Duration stats"
    )
    print_table(
        stats["returns"] * 100,
        float_format="{:.2f}%".format,
        name="Return stats",
    )

    if not hide_pos:
        stats["symbols"].columns = stats["symbols"].columns.map(format_asset)
        print_table(
            stats["symbols"] * 100,
            float_format="{:.2f}%".format,
            name="Symbol stats",
        )

def cal_mae_mfe(roundtrips:pd.DataFrame, days:int):
    """" 20230928 計算MAE, GMFE, BMFE等指標 Terry"""
    from tqdm import tqdm
    from zipline.data import bundles
    bundle = bundles.load('tquant')

    if days != None:
        roundtrips['prices'] = [[0] * days]*len(roundtrips)
    
    all_prices = []

    roundtrips['GMFE'], roundtrips['MAE'], roundtrips['BMFE'], roundtrips['MDD'] = 0,0,0,0

    rts = roundtrips.set_index(['symbol', 'open_dt', 'close_dt']).copy()

    for stock, st, et in tqdm(zip(rts.index.get_level_values(0), rts.index.get_level_values(1), rts.index.get_level_values(2)), total=len(rts)):

        tmp = get_prices(start_date=pd.Timestamp(st.date(), tz='utc'), 
                end_date=pd.Timestamp(et.date(), tz='utc'),
                field='close',
                assets=[stock])
        
        if days == None:
            all_prices.append(tmp[stock].tolist())
            
        try: 
            if days < len(tmp):
                tmp = tmp.head(days)
        except: pass

        rts.loc[(stock, st, et), 'returns_tot'] = tmp.iloc[-1].div(tmp.iloc[0]).sub(1).mul(100).values[0]

        rts.loc[(stock, st, et), 'GMFE'] = tmp.max().div(tmp.iloc[0]).sub(1).mul(100).values[0]

        rts.loc[(stock, st, et), 'MAE'] = tmp.min().div(tmp.iloc[0]).sub(1).mul(100).values[0]
        
        try: 
            rts.loc[(stock, st, et), 'BMFE'] = (tmp[tmp.index <= tmp.idxmin()[0]].max().div(tmp.iloc[0]).sub(1)).mul(100).values[0]
            
        except ValueError:
            rts.loc[(stock, st, et), 'BMFE'] = 0
        
        if days != None:
            rts.at[(stock, st, et), 'prices'] = tmp.iloc[:, 0].tolist()

        rts.loc[(stock, st, et), 'MDD'] = tmp.div(tmp.cummax()).sub(1).min().mul(100)[0]
    
    if days == None:
        rts = rts.assign(prices=all_prices)
        
    rts.reset_index(inplace=True)

    mae = rts[['symbol', 'returns_tot', 'MDD', 'GMFE', 'MAE', 'BMFE', 'prices']]

    return mae


from . import utils
from . import timeseries
from . import pos
from . import txn
from . import interesting_periods
from . import capacity
from . import round_trips
from . import perf_attrib

from .tears import *  # noqa
from .plotting import *  # noqa
from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

__all__ = [
    "utils",
    "timeseries",
    "pos",
    "txn",
    "interesting_periods",
    "capacity",
    "round_trips",
    "perf_attrib",
    "plot_cn_name_map"  #20231010 modify by yo
]



from parameterized import parameterized

from unittest import TestCase

from pandas import (
    Series,
    DataFrame,
    DatetimeIndex,
    date_range,
    Timedelta,
    read_csv,
)
from pandas.testing import assert_frame_equal

import os
import gzip

from pyfolio.round_trips import (
    extract_round_trips,
    add_closing_transactions,
    _groupby_consecutive,
)


class RoundTripTestCase(TestCase):
    dates = date_range(start="2015-01-01", freq="D", periods=20)
    dates_intraday = date_range(start="2015-01-01", freq="2BH", periods=8)

    @parameterized.expand(
        [
            (
                DataFrame(
                    data=[
                        [2, 10.0, "A"],
                        [2, 20.0, "A"],
                        [-2, 20.0, "A"],
                        [-2, 10.0, "A"],
                    ],
                    columns=["amount", "price", "symbol"],
                    index=dates_intraday[:4],
                ),
                DataFrame(
                    data=[
                        [4, 15.0, "A"],
                        [-4, 15.0, "A"],
                    ],
                    columns=["amount", "price", "symbol"],
                    index=dates_intraday[[0, 2]],
                ).rename_axis("dt", axis="index"),
            ),
            (
                DataFrame(
                    data=[
                        [2, 10.0, "A"],
                        [2, 20.0, "A"],
                        [2, 20.0, "A"],
                        [2, 10.0, "A"],
                    ],
                    columns=["amount", "price", "symbol"],
                    index=dates_intraday[[0, 1, 4, 5]],
                ),
                DataFrame(
                    data=[
                        [4, 15.0, "A"],
                        [4, 15.0, "A"],
                    ],
                    columns=["amount", "price", "symbol"],
                    index=dates_intraday[[0, 4]],
                ).rename_axis("dt", axis="index"),
            ),
        ]
    )
    def test_groupby_consecutive(self, transactions, expected):
        grouped_txn = _groupby_consecutive(transactions)
        assert_frame_equal(
            grouped_txn.sort_index(axis="columns"),
            expected.sort_index(axis="columns"),
        )

    @parameterized.expand(
        [
            # Simple round-trip
            (
                DataFrame(
                    data=[[2, 10.0, "A"], [-2, 15.0, "A"]],
                    columns=["amount", "price", "symbol"],
                    index=dates[:2],
                ),
                DataFrame(
                    data=[
                        [
                            dates[0],
                            dates[1],
                            Timedelta(days=1),
                            10.0,
                            0.5,
                            True,
                            "A",
                        ]
                    ],
                    columns=[
                        "open_dt",
                        "close_dt",
                        "duration",
                        "pnl",
                        "rt_returns",
                        "long",
                        "symbol",
                    ],
                    index=[0],
                ),
            ),
            # Round-trip with left-over txn that shouldn't be counted
            (
                DataFrame(
                    data=[[2, 10.0, "A"], [2, 15.0, "A"], [-9, 10.0, "A"]],
                    columns=["amount", "price", "symbol"],
                    index=dates[:3],
                ),
                DataFrame(
                    data=[
                        [
                            dates[0],
                            dates[2],
                            Timedelta(days=2),
                            -10.0,
                            -0.2,
                            True,
                            "A",
                        ]
                    ],
                    columns=[
                        "open_dt",
                        "close_dt",
                        "duration",
                        "pnl",
                        "rt_returns",
                        "long",
                        "symbol",
                    ],
                    index=[0],
                ),
            ),
            # Round-trip with sell that crosses 0 and should be split
            (
                DataFrame(
                    data=[[2, 10.0, "A"], [-4, 15.0, "A"], [3, 20.0, "A"]],
                    columns=["amount", "price", "symbol"],
                    index=dates[:3],
                ),
                DataFrame(
                    data=[
                        [
                            dates[0],
                            dates[1],
                            Timedelta(days=1),
                            10.0,
                            0.5,
                            True,
                            "A",
                        ],
                        [
                            dates[1],
                            dates[2],
                            Timedelta(days=1),
                            -10,
                            (-1.0 / 3),
                            False,
                            "A",
                        ],
                    ],
                    columns=[
                        "open_dt",
                        "close_dt",
                        "duration",
                        "pnl",
                        "rt_returns",
                        "long",
                        "symbol",
                    ],
                    index=[0, 1],
                ),
            ),
            # Round-trip that does not cross 0
            (
                DataFrame(
                    data=[[4, 10.0, "A"], [-2, 15.0, "A"], [2, 20.0, "A"]],
                    columns=["amount", "price", "symbol"],
                    index=dates[:3],
                ),
                DataFrame(
                    data=[
                        [
                            dates[0],
                            dates[1],
                            Timedelta(days=1),
                            10.0,
                            0.5,
                            True,
                            "A",
                        ]
                    ],
                    columns=[
                        "open_dt",
                        "close_dt",
                        "duration",
                        "pnl",
                        "rt_returns",
                        "long",
                        "symbol",
                    ],
                    index=[0],
                ),
            ),
            # Round-trip that does not cross 0 and has portfolio value
            (
                DataFrame(
                    data=[[4, 10.0, "A"], [-2, 15.0, "A"], [2, 20.0, "A"]],
                    columns=["amount", "price", "symbol"],
                    index=dates[:3],
                ),
                DataFrame(
                    data=[
                        [
                            dates[0],
                            dates[1],
                            Timedelta(days=1),
                            10.0,
                            0.5,
                            True,
                            "A",
                            0.1,
                        ]
                    ],
                    columns=[
                        "open_dt",
                        "close_dt",
                        "duration",
                        "pnl",
                        "rt_returns",
                        "long",
                        "symbol",
                        "returns",
                    ],
                    index=[0],
                ),
                Series([100.0, 100.0, 100.0], index=dates[:3]),
            ),
        ]
    )
    def test_extract_round_trips(
        self, transactions, expected, portfolio_value=None
    ):
        round_trips = extract_round_trips(
            transactions, portfolio_value=portfolio_value
        )

        assert_frame_equal(
            round_trips.sort_index(axis="columns"),
            expected.sort_index(axis="columns"),
        )

    def test_add_closing_trades(self):
        dates = date_range(start="2015-01-01", periods=20)
        transactions = DataFrame(
            data=[[2, 10, "A"], [-5, 10, "A"], [-1, 10, "B"]],
            columns=["amount", "price", "symbol"],
            index=dates[:3],
        )
        positions = DataFrame(
            data=[[20, 10, 0], [-30, 10, 30], [-60, 0, 30]],
            columns=["A", "B", "cash"],
            index=dates[:3],
        )

        expected_ix = dates[:3].append(
            DatetimeIndex([dates[2] + Timedelta(seconds=1)])
        )
        expected = DataFrame(
            data=[
                ["A", 2, 10],
                ["A", -5, 10],
                ["B", -1, 10.0],
                [
                    "A",
                    3,
                    20.0,
                ],
            ],
            columns=["symbol", "amount", "price"],
            index=expected_ix,
        )

        transactions_closed = add_closing_transactions(positions, transactions)
        assert_frame_equal(transactions_closed, expected)

    def test_txn_pnl_matches_round_trip_pnl(self):
        __location__ = os.path.realpath(
            os.path.join(os.getcwd(), os.path.dirname(__file__))
        )

        test_txn = read_csv(
            gzip.open(__location__ + "/test_data/test_txn.csv.gz"),
            index_col=0,
            parse_dates=True,
        )
        test_pos = read_csv(
            gzip.open(__location__ + "/test_data/test_pos.csv.gz"),
            index_col=0,
            parse_dates=True,
        )

        transactions_closed = add_closing_transactions(test_pos, test_txn)
        transactions_closed["txn_dollars"] = (
            transactions_closed.amount * -1.0 * transactions_closed.price
        )
        round_trips = extract_round_trips(transactions_closed)

        self.assertAlmostEqual(
            round_trips.pnl.sum(), transactions_closed.txn_dollars.sum()
        )



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from itertools import cycle
from matplotlib.pyplot import cm
import numpy as np
import pandas as pd
from IPython.display import display, HTML
from distutils.version import StrictVersion
import empyrical.utils
import inspect
import re
import pytz

from . import pos
from . import txn

APPROX_BDAYS_PER_MONTH = 21
APPROX_BDAYS_PER_YEAR = 252

MONTHS_PER_YEAR = 12
WEEKS_PER_YEAR = 52

MM_DISPLAY_UNIT = 1000000.0

DAILY = "daily"
WEEKLY = "weekly"
MONTHLY = "monthly"
YEARLY = "yearly"

ANNUALIZATION_FACTORS = {
    DAILY: APPROX_BDAYS_PER_YEAR,
    WEEKLY: WEEKS_PER_YEAR,
    MONTHLY: MONTHS_PER_YEAR,
}

COLORMAP = "Paired"
COLORS = [
    "#e6194b",
    "#3cb44b",
    "#ffe119",
    "#0082c8",
    "#f58231",
    "#911eb4",
    "#46f0f0",
    "#f032e6",
    "#d2f53c",
    "#fabebe",
    "#008080",
    "#e6beff",
    "#aa6e28",
    "#800000",
    "#aaffc3",
    "#808000",
    "#ffd8b1",
    "#000080",
    "#808080",
]

pandas_version = StrictVersion(pd.__version__)

pandas_one_point_one_or_less = pandas_version < StrictVersion("1.2")


def one_dec_places(x, pos):
    """
    Adds 1/10th decimal to plot ticks.
    """

    return "%.1f" % x


def two_dec_places(x, pos):
    """
    Adds 1/100th decimal to plot ticks.
    """

    return "%.2f" % x


def percentage(x, pos):
    """
    Adds percentage sign to plot ticks.
    """

    return "%.0f%%" % x


def format_asset(asset):
    """
    If zipline asset objects are used, we want to print them out prettily
    within the tear sheet. This function should only be applied directly
    before displaying.
    """

    try:
        import zipline.assets
    except ImportError:
        return asset

    if isinstance(asset, zipline.assets.Asset):
        return asset.symbol
    else:
        return asset


def vectorize(func):
    """
    Decorator so that functions can be written to work on Series but
    may still be called with DataFrames.
    """

    def wrapper(df, *args, **kwargs):
        if df.ndim == 1:
            return func(df, *args, **kwargs)
        elif df.ndim == 2:
            return df.apply(func, *args, **kwargs)

    return wrapper


def extract_rets_pos_txn_from_zipline(backtest):
    """
    Extract returns, positions, transactions and leverage from the
    backtest data structure returned by zipline.TradingAlgorithm.run().

    The returned data structures are in a format compatible with the
    rest of pyfolio and can be directly passed to
    e.g. tears.create_full_tear_sheet().

    Parameters
    ----------
    backtest : pd.DataFrame
        DataFrame returned by zipline.TradingAlgorithm.run()

    Returns
    -------
    returns : pd.Series
        Daily returns of strategy.
         - See full explanation in tears.create_full_tear_sheet.
    positions : pd.DataFrame
        Daily net position values.
         - See full explanation in tears.create_full_tear_sheet.
    transactions : pd.DataFrame
        Prices and amounts of executed trades. One row per trade.
         - See full explanation in tears.create_full_tear_sheet.


    Example (on the Quantopian research platform)
    ---------------------------------------------
    >>> backtest = my_algo.run()
    >>> returns, positions, transactions =
    >>>     pyfolio.utils.extract_rets_pos_txn_from_zipline(backtest)
    >>> pyfolio.tears.create_full_tear_sheet(returns,
    >>>     positions, transactions)
    """
    backtest.index = backtest.index.normalize()
    if backtest.index.tzinfo is None:
        backtest.index = backtest.index.tz_localize("UTC")
    # The `positions` and `returns` data must have a tz-aware DateTimeIndex set to UTC, with a time 
    # of 0:00, otherwise some plots won't be able to be generated. (20231031)
    elif backtest.index.tzinfo!=pytz.utc:
        backtest.index = backtest.index.tz_localize(None).tz_localize('UTC')

    returns = backtest.returns

    raw_positions = []
    for dt, pos_row in backtest.positions.iteritems():
        df = pd.DataFrame(pos_row)
        df.index = [dt] * len(df)
        raw_positions.append(df)
    if not raw_positions:
        raise ValueError("The backtest does not have any positions.")
    positions = pd.concat(raw_positions)
    positions = pos.extract_pos(positions, backtest.ending_cash)

    transactions = txn.make_transaction_frame(backtest.transactions)
    if transactions.index.tzinfo is None:
        transactions.index = transactions.index.tz_localize("utc")
    # The `transactions`` data must have a tz-aware DateTimeIndex set to UTC, with a time 
    # of 0:00, otherwise some plots won't be able to be generated. (20231031)
    elif transactions.index.tzinfo!=pytz.utc:
        transactions.index = transactions.index.tz_localize(None).tz_localize('UTC')

    return returns, positions, transactions




def _is_running_in_spyder():
    """
    The `_is_running_in_spyder` function determines if the current script
    is being executed within the "Spyder" Integrated Development Environment
    (IDE). It inspects the execution stack to search for any trace of "spyder"
    in the "file paths". This check is necessary because Spyder seems to have issues
    displaying tables using the `to_html` method. By identifying if the script runs in
    Spyder, alternative table rendering methods can be employed to avoid display issues.
    (20230908)
    """
    for f in inspect.stack():
        # print(f.filename)
        if re.search("spyder",f.filename):
            return True
    return False


def print_table(
    table, name=None, float_format=None, formatters=None, header_rows=None
):
    """
    Pretty print a pandas DataFrame.

    Uses HTML output if running inside Jupyter Notebook, otherwise
    formatted text output.

    Parameters
    ----------
    table : pandas.Series or pandas.DataFrame
        Table to pretty-print.
    name : str, optional
        Table name to display in upper left corner.
    float_format : function, optional
        Formatter to use for displaying table elements, passed as the
        `float_format` arg to pd.Dataframe.to_html.
        E.g. `'{0:.2%}'.format` for displaying 100 as '100.00%'.
    formatters : list or dict, optional
        Formatters to use by column, passed as the `formatters` arg to
        pd.Dataframe.to_html.
    header_rows : dict, optional
        Extra rows to display at the top of the table.
    """

    if isinstance(table, pd.Series):
        table = pd.DataFrame(table)

    if name is not None:
        table.columns.name = name

    # Spyder seems to have issues displaying tables using the `to_html` method.
    # By identifying whether the script is running in Spyder, we can use alternative
    # table rendering methods instead.(20230908)
    if not _is_running_in_spyder():
        html = table.to_html(float_format=float_format, formatters=formatters)

        if header_rows is not None:
            # Count the number of columns for the text to span
            n_cols = html.split("<thead>")[1].split("</thead>")[0].count("<th>")

            # Generate the HTML for the extra rows
            rows = ""
            for name, value in header_rows.items():
                rows += (
                    '\n    <tr style="text-align: right;"><th>%s</th>'
                    + "<td colspan=%d>%s</td></tr>"
                ) % (name, n_cols, value)

            # Inject the new HTML
            html = html.replace("<thead>", "<thead>" + rows)

        display(HTML(html))

    else:
        print(table)


def standardize_data(x):
    """
    Standardize an array with mean and standard deviation.

    Parameters
    ----------
    x : np.array
        Array to standardize.

    Returns
    -------
    np.array
        Standardized array.
    """

    return (x - np.mean(x)) / np.std(x)


def detect_intraday(positions, transactions, threshold=0.25):
    """
    Attempt to detect an intraday strategy. Get the number of
    positions held at the end of the day, and divide that by the
    number of unique stocks transacted every day. If the average quotient
    is below a threshold, then an intraday strategy is detected.

    Parameters
    ----------
    positions : pd.DataFrame
        Daily net position values.
         - See full explanation in create_full_tear_sheet.
    transactions : pd.DataFrame
        Prices and amounts of executed trades. One row per trade.
         - See full explanation in create_full_tear_sheet.

    Returns
    -------
    boolean
        True if an intraday strategy is detected.
    """

    daily_txn = transactions.copy()
    daily_txn.index = daily_txn.index.date
    txn_count = daily_txn.groupby(level=0).symbol.nunique().sum()
    daily_pos = positions.drop("cash", axis=1).replace(0, np.nan)
    return daily_pos.count(axis=1).sum() / txn_count < threshold


def check_intraday(estimate, returns, positions, transactions):
    """
    Logic for checking if a strategy is intraday and processing it.

    Parameters
    ----------
    estimate: boolean or str, optional
        Approximate returns for intraday strategies.
        See description in tears.create_full_tear_sheet.
    returns : pd.Series
        Daily returns of the strategy, noncumulative.
         - See full explanation in create_full_tear_sheet.
    positions : pd.DataFrame
        Daily net position values.
         - See full explanation in create_full_tear_sheet.
    transactions : pd.DataFrame
        Prices and amounts of executed trades. One row per trade.
         - See full explanation in create_full_tear_sheet.

    Returns
    -------
    pd.DataFrame
        Daily net position values, adjusted for intraday movement.
    """

    if estimate == "infer":
        if positions is not None and transactions is not None:
            if detect_intraday(positions, transactions):
                warnings.warn(
                    "Detected intraday strategy; inferring positi"
                    + "ons from transactions. Set estimate_intraday"
                    + "=False to disable."
                )
                return estimate_intraday(returns, positions, transactions)
            else:
                return positions
        else:
            return positions

    elif estimate:
        if positions is not None and transactions is not None:
            return estimate_intraday(returns, positions, transactions)
        else:
            raise ValueError("Positions and txns needed to estimate intraday")
    else:
        return positions


def estimate_intraday(returns, positions, transactions, EOD_hour=23):
    """
    Intraday strategies will often not hold positions at the day end.
    This attempts to find the point in the day that best represents
    the activity of the strategy on that day, and effectively resamples
    the end-of-day positions with the positions at this point of day.
    The point of day is found by detecting when our exposure in the
    market is at its maximum point. Note that this is an estimate.

    Parameters
    ----------
    returns : pd.Series
        Daily returns of the strategy, noncumulative.
         - See full explanation in create_full_tear_sheet.
    positions : pd.DataFrame
        Daily net position values.
         - See full explanation in create_full_tear_sheet.
    transactions : pd.DataFrame
        Prices and amounts of executed trades. One row per trade.
         - See full explanation in create_full_tear_sheet.

    Returns
    -------
    pd.DataFrame
        Daily net position values, resampled for intraday behavior.
    """

    # Construct DataFrame of transaction amounts
    txn_val = transactions.copy()
    txn_val.index.names = ["date"]
    txn_val["value"] = txn_val.amount * txn_val.price
    txn_val = (
        txn_val.reset_index()
        .pivot_table(index="date", values="value", columns="symbol")
        .replace(np.nan, 0)
    )

    # Cumulate transaction amounts each day
    txn_val = txn_val.groupby(txn_val.index.date).cumsum()

    # Calculate exposure, then take peak of exposure every day
    txn_val["exposure"] = txn_val.abs().sum(axis=1)
    condition = txn_val["exposure"] == txn_val.groupby(pd.Grouper(freq="24H"))[
        "exposure"
    ].transform(max)
    txn_val = txn_val[condition].drop("exposure", axis=1)

    # Compute cash delta
    txn_val["cash"] = -txn_val.sum(axis=1)

    # Shift EOD positions to positions at start of next trading day
    positions_shifted = positions.copy().shift(1).fillna(0)
    starting_capital = positions.iloc[0].sum() / (1 + returns[0])
    positions_shifted.cash[0] = starting_capital

    # Format and add start positions to intraday position changes
    txn_val.index = txn_val.index.normalize()
    corrected_positions = positions_shifted.add(txn_val, fill_value=0)
    corrected_positions.index.name = "period_close"
    corrected_positions.columns.name = "sid"

    return corrected_positions


def clip_returns_to_benchmark(rets, benchmark_rets):
    """
    Drop entries from rets so that the start and end dates of rets match those
    of benchmark_rets.

    Parameters
    ----------
    rets : pd.Series
        Daily returns of the strategy, noncumulative.
         - See pf.tears.create_full_tear_sheet for more details

    benchmark_rets : pd.Series
        Daily returns of the benchmark, noncumulative.

    Returns
    -------
    clipped_rets : pd.Series
        Daily noncumulative returns with index clipped to match that of
        benchmark returns.
    """

    if (rets.index[0] < benchmark_rets.index[0]) or (
        rets.index[-1] > benchmark_rets.index[-1]
    ):
        clipped_rets = rets[benchmark_rets.index]
    else:
        clipped_rets = rets

    return clipped_rets


def to_utc(df):
    """
    For use in tests; applied UTC timestamp to DataFrame.
    """

    try:
        df.index = df.index.tz_localize("UTC")
    except TypeError:
        df.index = df.index.tz_convert("UTC")

    return df


def to_series(df):
    """
    For use in tests; converts DataFrame's first column to Series.
    """

    return df[df.columns[0]]


# This functions is simply a passthrough to empyrical, but is
# required by the register_returns_func and get_symbol_rets.
default_returns_func = empyrical.utils.default_returns_func

# Settings dict to store functions/values that may
# need to be overridden depending on the users environment
SETTINGS = {"returns_func": default_returns_func}


def register_return_func(func):
    """
    Registers the 'returns_func' that will be called for
    retrieving returns data.

    Parameters
    ----------
    func : function
        A function that returns a pandas Series of asset returns.
        The signature of the function must be as follows

        >>> func(symbol)

        Where symbol is an asset identifier

    Returns
    -------
    None
    """

    SETTINGS["returns_func"] = func


def get_symbol_rets(symbol, start=None, end=None):
    """
    Calls the currently registered 'returns_func'

    Parameters
    ----------
    symbol : object
        An identifier for the asset whose return
        series is desired.
        e.g. ticker symbol or database ID
    start : date, optional
        Earliest date to fetch data for.
        Defaults to earliest date available.
    end : date, optional
        Latest date to fetch data for.
        Defaults to latest date available.

    Returns
    -------
    pandas.Series
        Returned by the current 'returns_func'
    """

    return SETTINGS["returns_func"](symbol, start=start, end=end)


def configure_legend(
    ax, autofmt_xdate=True, change_colors=False, rotation=30, ha="right"
):
    """
    Format legend for perf attribution plots:
    - put legend to the right of plot instead of overlapping with it
    - make legend order match up with graph lines
    - set colors according to colormap
    """
    chartBox = ax.get_position()
    ax.set_position(
        [chartBox.x0, chartBox.y0, chartBox.width * 0.75, chartBox.height]
    )

    # make legend order match graph lines
    handles, labels = ax.get_legend_handles_labels()
    handles_and_labels_sorted = sorted(
        zip(handles, labels), key=lambda x: x[0].get_ydata()[-1], reverse=True
    )

    handles_sorted = [h[0] for h in handles_and_labels_sorted]
    labels_sorted = [h[1] for h in handles_and_labels_sorted]

    if change_colors:
        for handle, color in zip(handles_sorted, cycle(COLORS)):
            handle.set_color(color)

    ax.legend(
        handles=handles_sorted,
        labels=labels_sorted,
        frameon=True,
        framealpha=0.5,
        loc="upper left",
        bbox_to_anchor=(1.05, 1),
        fontsize="small",
    )

    # manually rotate xticklabels instead of using matplotlib's autofmt_xdate
    # because it disables xticklabels for all but the last plot
    if autofmt_xdate:
        for label in ax.get_xticklabels():
            label.set_ha(ha)
            label.set_rotation(rotation)


def sample_colormap(cmap_name, n_samples):
    """
    Sample a colormap from matplotlib
    """
    colors = []
    colormap = cm.cmap_d[cmap_name]
    for i in np.linspace(0, 1, n_samples):
        colors.append(colormap(i))

    return colors



import empyrical as ep
import numpy as np
import pandas as pd

from . import pos


def daily_txns_with_bar_data(transactions, market_data):
    """
    Sums the absolute value of shares traded in each name on each day.
    Adds columns containing the closing price and total daily volume(from market_data not transactions)
    for each day-ticker combination.

    Parameters
    ----------
    transactions : pd.DataFrame
        Prices and amounts of executed trades. One row per trade.
        - See full explanation in tears.create_full_tear_sheet
    market_data : pd.DataFrame
        Daily market_data
        - DataFrame has a multi-index index, one level is dates and another is
        market_data contains volume & price, equities as columns

    Returns
    -------
    txn_daily : pd.DataFrame
        Daily totals for transacted shares in each traded name.
        price and volume columns for close price and daily volume for
        the corresponding ticker, respectively.
    """

    transactions.index.name = "date"
    txn_daily = pd.DataFrame(
        transactions.assign(amount=abs(transactions.amount))
        .groupby(["symbol", pd.Grouper(freq="D")])  
        .sum()["amount"]
    )  # transactions的日期一定是交易日，pd.Grouper(freq="D")理應不會取到非交易日。

    txn_daily["price"] = market_data.xs("price", level=1).unstack()
    txn_daily["volume"] = market_data.xs("volume", level=1).unstack() 

    txn_daily = (
        txn_daily.reset_index().set_index("date").sort_index()#.asfreq("D") 
    ) 
    #20230726 (by MRC) 
    #刪除.asfreq("D")，不然會報錯："ValueError: cannot reindex from a duplicate axis"
    #前面txn_daily也已經Grouper(freq="D")了，所以理應不用.asfreq("D")
  
    return txn_daily


def days_to_liquidate_positions(
    positions,
    market_data,
    max_bar_consumption=0.2,
    capital_base=1e6,
    mean_volume_window=5,
):
    """
    Compute the number of days that would have been required
    to fully liquidate each position on each day based on the
    trailing n day mean daily bar volume and a limit on the proportion
    of a daily bar that we are allowed to consume.

    This analysis uses portfolio allocations and a provided capital base
    rather than the dollar values in the positions DataFrame to remove the
    effect of compounding on days to liquidate. In other words, this function
    assumes that the net liquidation portfolio value will always remain
    constant at capital_base.

    Parameters
    ----------
    positions: pd.DataFrame
        Contains daily position values including cash
        - See full explanation in tears.create_full_tear_sheet
    market_data : pd.DataFrame
        Daily market_data
        - DataFrame has a multi-index index, one level is dates and another is
        market_data contains volume & price, equities as columns
    max_bar_consumption : float
        Max proportion of a daily bar that can be consumed in the
        process of liquidating a position.
    capital_base : integer
        Capital base multiplied by portfolio allocation to compute
        position value that needs liquidating.
    mean_volume_window : float
        Trailing window to use in mean volume calculation.

    Returns
    -------
    days_to_liquidate : pd.DataFrame
        Number of days required to fully liquidate daily positions.
        Datetime index, symbols as columns.
    """

    DV = market_data.xs("volume", level=1) * market_data.xs("price", level=1)
    roll_mean_dv = (
        DV.rolling(window=mean_volume_window, center=False).mean().shift()
    )
    roll_mean_dv = roll_mean_dv.replace(0, np.nan)

    positions_alloc = pos.get_percent_alloc(positions)
    positions_alloc = positions_alloc.drop("cash", axis=1)

    days_to_liquidate = (positions_alloc * capital_base) / (
        max_bar_consumption * roll_mean_dv
    )

    return days_to_liquidate.iloc[mean_volume_window:]


def get_max_days_to_liquidate_by_ticker(
    positions,
    market_data,
    max_bar_consumption=0.2,
    capital_base=1e6,
    mean_volume_window=5,
    last_n_days=None,
):
    """
    Finds the longest estimated liquidation time for each traded
    name over the course of backtest (or last n days of the backtest).

    Parameters
    ----------
    positions: pd.DataFrame
        Contains daily position values including cash
        - See full explanation in tears.create_full_tear_sheet
    market_data : pd.DataFrame
        Daily market_data
        - DataFrame has a multi-index index, one level is dates and another is
        market_data contains volume & price, equities as columns
    max_bar_consumption : float
        Max proportion of a daily bar that can be consumed in the
        process of liquidating a position.
    capital_base : integer
        Capital base multiplied by portfolio allocation to compute
        position value that needs liquidating.
    mean_volume_window : float
        Trailing window to use in mean volume calculation.
    last_n_days : integer
        Compute for only the last n days of the passed backtest data.

    Returns
    -------
    days_to_liquidate : pd.DataFrame
        Max Number of days required to fully liquidate each traded name.
        Index of symbols. Columns for days_to_liquidate and the corresponding
        date and position_alloc on that day.
    """

    dtlp = days_to_liquidate_positions(
        positions,
        market_data,
        max_bar_consumption=max_bar_consumption,
        capital_base=capital_base,
        mean_volume_window=mean_volume_window,
    )

    if last_n_days is not None:
        #dtlp = dtlp.loc[dtlp.index.max() - pd.Timedelta(days=last_n_days) :] #日曆日
        dtlp.loc[np.unique(market_data.index.get_level_values(0).date)[-last_n_days-1]:]
        #20230726 (by MRC) 日曆日->交易日，但當market_data並非每個交易日都有的時候會有問題。

    pos_alloc = pos.get_percent_alloc(positions)
    pos_alloc = pos_alloc.drop("cash", axis=1)

    liq_desc = pd.DataFrame()
    liq_desc["days_to_liquidate"] = dtlp.unstack()
    liq_desc["pos_alloc_pct"] = pos_alloc.unstack() * 100
    liq_desc.index.set_names(["symbol", "date"], inplace=True)

    worst_liq = (
        liq_desc.reset_index()
        .sort_values("days_to_liquidate", ascending=False) 
        .groupby("symbol")
        .first()
    )   # 每一個symbol取days_to_liquidate最大的那筆。

    return worst_liq


def get_low_liquidity_transactions(
    transactions, market_data, last_n_days=None
):
    """
    For each traded name, find the daily transaction total that consumed
    the greatest proportion of available daily bar volume.

    Parameters
    ----------
    transactions : pd.DataFrame
        Prices and amounts of executed trades. One row per trade.
         - See full explanation in create_full_tear_sheet.
    market_data : pd.DataFrame
        Daily market_data
        - DataFrame has a multi-index index, one level is dates and another is
        market_data contains volume & price, equities as columns
    last_n_days : integer
        Compute for only the last n days of the passed backtest data.
    """

    txn_daily_w_bar = daily_txns_with_bar_data(transactions, market_data)
    txn_daily_w_bar.index.name = "date"

    if last_n_days is not None:
        #md = txn_daily_w_bar.date.max() - pd.Timedelta(days=last_n_days)   # 日曆日
        #txn_daily_w_bar = txn_daily_w_bar[txn_daily_w_bar.date > md]       # 日曆日
        txn_daily_w_bar = txn_daily_w_bar.loc[np.unique(market_data.index.get_level_values(0).date)[-last_n_days-1]:]
        #20230726 (by MRC) 日曆日->交易日，當market_data並非每個交易日都有的時候會有問題。

    txn_daily_w_bar = txn_daily_w_bar.reset_index()   #20230726 (by MRC) 搭配以上日曆日轉交易日的更改

    bar_consumption = txn_daily_w_bar.assign(
        max_pct_bar_consumed=txn_daily_w_bar.amount.div(
            txn_daily_w_bar.volume
        ).mul(100)
    ).sort_values("max_pct_bar_consumed", ascending=False)
    max_bar_consumption = bar_consumption.groupby("symbol").first()

    return max_bar_consumption[["date", "max_pct_bar_consumed"]]


def apply_slippage_penalty(
    returns,
    txn_daily,
    simulate_starting_capital,
    backtest_starting_capital,
    impact=0.1,
):
    """
    Applies quadratic volumeshare slippage model to daily returns based
    on the proportion of the observed historical daily bar dollar volume
    consumed by the strategy's trades. Scales the size of trades based
    on the ratio of the starting capital we wish to test to the starting
    capital of the passed backtest data.

    Parameters
    ----------
    returns : pd.Series
        Time series of daily returns.
    txn_daily : pd.Series
        Daily transaciton totals, closing price, and daily volume for
        each traded name. See price_volume_daily_txns for more details.
    simulate_starting_capital : integer
        capital at which we want to test
    backtest_starting_capital: capital base at which backtest was
        origionally run. impact: See Zipline volumeshare slippage model
    impact : float
        Scales the size of the slippage penalty.

    Returns
    -------
    adj_returns : pd.Series
        Slippage penalty adjusted daily returns.
    """

    mult = simulate_starting_capital / backtest_starting_capital
    simulate_traded_shares = abs(mult * txn_daily.amount)
    simulate_traded_dollars = txn_daily.price * simulate_traded_shares
    simulate_pct_volume_used = simulate_traded_shares / txn_daily.volume

    penalties = (
        simulate_pct_volume_used ** 2 * impact * simulate_traded_dollars
    )

    daily_penalty = penalties.resample("D").sum()
    daily_penalty = daily_penalty.reindex(returns.index).fillna(0)

    # Since we are scaling the numerator of the penalties linearly
    # by capital base, it makes the most sense to scale the denominator
    # similarly. In other words, since we aren't applying compounding to
    # simulate_traded_shares, we shouldn't apply compounding to pv.
    portfolio_value = (
        ep.cum_returns(returns, starting_value=backtest_starting_capital)
        * mult
    )

    adj_returns = returns - (daily_penalty / portfolio_value)

    return adj_returns



import os
from unittest import TestCase, skipIf
from parameterized import parameterized
from numpy.testing import assert_allclose, assert_almost_equal
from pandas.testing import assert_series_equal
import numpy as np
import pandas as pd

from .. import timeseries
from pyfolio.utils import to_utc, to_series, pandas_one_point_one_or_less

import gzip

DECIMAL_PLACES = 8


class TestDrawdown(TestCase):
    drawdown_list = np.array([100, 90, 75]) / 10.0
    dt = pd.date_range("2000-1-3", periods=3, freq="D")

    drawdown_serie = pd.Series(drawdown_list, index=dt)

    @parameterized.expand([(drawdown_serie,)])
    def test_get_max_drawdown_begins_first_day(self, px):
        rets = px.pct_change()
        drawdowns = timeseries.gen_drawdown_table(rets, top=1)
        self.assertEqual(drawdowns.loc[0, "Net drawdown in %"], 25)

    drawdown_list = (
        np.array(
            [
                100,
                110,
                120,
                150,
                180,
                200,
                100,
                120,
                160,
                180,
                200,
                300,
                400,
                500,
                600,
                800,
                900,
                1000,
                650,
                600,
            ]
        )
        / 10.0
    )
    dt = pd.date_range("2000-1-3", periods=20, freq="D")

    drawdown_serie = pd.Series(drawdown_list, index=dt)

    @parameterized.expand(
        [
            (
                drawdown_serie,
                pd.Timestamp("2000-01-08"),
                pd.Timestamp("2000-01-09"),
                pd.Timestamp("2000-01-13"),
                50,
                pd.Timestamp("2000-01-20"),
                pd.Timestamp("2000-01-22"),
                None,
                40,
            )
        ]
    )
    def test_gen_drawdown_table_relative(
        self,
        px,
        first_expected_peak,
        first_expected_valley,
        first_expected_recovery,
        first_net_drawdown,
        second_expected_peak,
        second_expected_valley,
        second_expected_recovery,
        second_net_drawdown,
    ):

        rets = px.pct_change()

        drawdowns = timeseries.gen_drawdown_table(rets, top=2)

        self.assertEqual(
            np.round(drawdowns.loc[0, "Net drawdown in %"]), first_net_drawdown
        )
        self.assertEqual(drawdowns.loc[0, "Peak date"], first_expected_peak)
        self.assertEqual(
            drawdowns.loc[0, "Valley date"], first_expected_valley
        )
        self.assertEqual(
            drawdowns.loc[0, "Recovery date"], first_expected_recovery
        )

        self.assertEqual(
            np.round(drawdowns.loc[1, "Net drawdown in %"]),
            second_net_drawdown,
        )
        self.assertEqual(drawdowns.loc[1, "Peak date"], second_expected_peak)
        self.assertEqual(
            drawdowns.loc[1, "Valley date"], second_expected_valley
        )
        self.assertTrue(pd.isnull(drawdowns.loc[1, "Recovery date"]))

    px_list_1 = (
        np.array([100, 120, 100, 80, 70, 110, 180, 150]) / 100.0
    )  # Simple
    px_list_2 = (
        np.array([100, 120, 100, 80, 70, 80, 90, 90]) / 100.0
    )  # Ends in drawdown
    dt = pd.date_range("2000-1-3", periods=8, freq="D")

    @parameterized.expand(
        [
            (
                pd.Series(px_list_1, index=dt),
                pd.Timestamp("2000-1-4"),
                pd.Timestamp("2000-1-7"),
                pd.Timestamp("2000-1-9"),
            ),
            (
                pd.Series(px_list_2, index=dt),
                pd.Timestamp("2000-1-4"),
                pd.Timestamp("2000-1-7"),
                None,
            ),
        ]
    )
    def test_get_max_drawdown(
        self, px, expected_peak, expected_valley, expected_recovery
    ):
        rets = px.pct_change().iloc[1:]

        peak, valley, recovery = timeseries.get_max_drawdown(rets)
        # Need to use isnull because the result can be NaN, NaT, etc.
        self.assertTrue(
            pd.isnull(peak)
        ) if expected_peak is None else self.assertEqual(peak, expected_peak)
        self.assertTrue(
            pd.isnull(valley)
        ) if expected_valley is None else self.assertEqual(
            valley, expected_valley
        )
        self.assertTrue(
            pd.isnull(recovery)
        ) if expected_recovery is None else self.assertEqual(
            recovery, expected_recovery
        )

    @parameterized.expand(
        [
            (
                pd.Series(px_list_2, index=dt),
                pd.Timestamp("2000-1-4"),
                pd.Timestamp("2000-1-7"),
                None,
                None,
            ),
            (
                pd.Series(px_list_1, index=dt),
                pd.Timestamp("2000-1-4"),
                pd.Timestamp("2000-1-7"),
                pd.Timestamp("2000-1-9"),
                4,
            ),
        ]
    )
    def test_gen_drawdown_table(
        self,
        px,
        expected_peak,
        expected_valley,
        expected_recovery,
        expected_duration,
    ):
        rets = px.pct_change().iloc[1:]

        drawdowns = timeseries.gen_drawdown_table(rets, top=1)
        self.assertTrue(
            pd.isnull(drawdowns.loc[0, "Peak date"])
        ) if expected_peak is None else self.assertEqual(
            drawdowns.loc[0, "Peak date"], expected_peak
        )
        self.assertTrue(
            pd.isnull(drawdowns.loc[0, "Valley date"])
        ) if expected_valley is None else self.assertEqual(
            drawdowns.loc[0, "Valley date"], expected_valley
        )
        self.assertTrue(
            pd.isnull(drawdowns.loc[0, "Recovery date"])
        ) if expected_recovery is None else self.assertEqual(
            drawdowns.loc[0, "Recovery date"], expected_recovery
        )
        self.assertTrue(
            pd.isnull(drawdowns.loc[0, "Duration"])
        ) if expected_duration is None else self.assertEqual(
            drawdowns.loc[0, "Duration"], expected_duration
        )

    def test_drawdown_overlaps(self):
        rand = np.random.RandomState(1337)
        n_samples = 252 * 5
        spy_returns = pd.Series(
            rand.standard_t(3.1, n_samples),
            pd.date_range("2005-01-02", periods=n_samples),
        )
        spy_drawdowns = timeseries.gen_drawdown_table(
            spy_returns, top=20
        ).sort_values(by="Peak date")
        # Compare the recovery date of each drawdown with the peak of the next
        # Last pair might contain a NaT if drawdown didn't finish, so ignore it
        pairs = list(
            zip(
                spy_drawdowns["Recovery date"],
                spy_drawdowns["Peak date"].shift(-1),
            )
        )[:-1]
        self.assertGreater(len(pairs), 0)
        for recovery, peak in pairs:
            if not pd.isnull(recovery):
                self.assertLessEqual(recovery, peak)

    @parameterized.expand(
        [
            (
                pd.Series(px_list_1, index=dt),
                1,
                [
                    (
                        pd.Timestamp("2000-01-03 00:00:00"),
                        pd.Timestamp("2000-01-03 00:00:00"),
                        pd.Timestamp("2000-01-03 00:00:00"),
                    )
                ],
            )
        ]
    )
    def test_top_drawdowns(self, returns, top, expected):
        self.assertEqual(
            timeseries.get_top_drawdowns(returns, top=top), expected
        )


class TestVariance(TestCase):
    @parameterized.expand([(1e7, 0.5, 1, 1, -10000000.0)])
    def test_var_cov_var_normal(self, P, c, mu, sigma, expected):
        self.assertEqual(
            timeseries.var_cov_var_normal(P, c, mu, sigma), expected
        )


class TestNormalize(TestCase):
    dt = pd.date_range("2000-1-3", periods=8, freq="D")
    px_list = [1.0, 1.2, 1.0, 0.8, 0.7, 0.8, 0.8, 0.8]

    @parameterized.expand(
        [
            (
                pd.Series(np.array(px_list) * 100, index=dt),
                pd.Series(px_list, index=dt),
            )
        ]
    )
    def test_normalize(self, returns, expected):
        self.assertTrue(timeseries.normalize(returns).equals(expected))


class TestStats(TestCase):
    simple_rets = pd.Series(
        [0.1] * 3 + [0] * 497, pd.date_range("2000-1-3", periods=500, freq="D")
    )

    simple_week_rets = pd.Series(
        [0.1] * 3 + [0] * 497,
        pd.date_range("2000-1-31", periods=500, freq="W"),
    )

    simple_month_rets = pd.Series(
        [0.1] * 3 + [0] * 497,
        pd.date_range("2000-1-31", periods=500, freq="M"),
    )

    simple_benchmark = pd.Series(
        [0.03] * 4 + [0] * 496,
        pd.date_range("2000-1-1", periods=500, freq="D"),
    )
    px_list = np.array([10, -10, 10]) / 100.0  # Ends in drawdown
    dt = pd.date_range("2000-1-3", periods=3, freq="D")

    px_list_2 = [1.0, 1.2, 1.0, 0.8, 0.7, 0.8, 0.8, 0.8]
    dt_2 = pd.date_range("2000-1-3", periods=8, freq="D")

    @parameterized.expand(
        [
            (
                simple_rets[:5],
                2,
                [np.nan, np.inf, np.inf, 11.224972160321, np.nan],
            )
        ]
    )
    @skipIf(
        pandas_one_point_one_or_less, "pandas<1.2 returns np.inf not np.nan"
    )
    def test_sharpe_2(self, returns, rolling_sharpe_window, expected):
        np.testing.assert_array_almost_equal(
            timeseries.rolling_sharpe(
                returns, rolling_sharpe_window
            ).to_numpy(),
            np.asarray(expected),
        )

    @parameterized.expand([(simple_rets[:5], simple_benchmark, 2, 0)])
    def test_beta(self, returns, benchmark_rets, rolling_window, expected):
        actual = timeseries.rolling_beta(
            returns,
            benchmark_rets,
            rolling_window=rolling_window,
        ).values.tolist()[2]

        np.testing.assert_almost_equal(actual, expected)


class TestCone(TestCase):
    def test_bootstrap_cone_against_linear_cone_normal_returns(self):
        random_seed = 100
        np.random.seed(random_seed)
        days_forward = 200
        cone_stdevs = (1.0, 1.5, 2.0)
        mu = 0.005
        sigma = 0.002
        rets = pd.Series(np.random.normal(mu, sigma, 10000))

        midline = np.cumprod(1 + (rets.mean() * np.ones(days_forward)))
        stdev = rets.std() * midline * np.sqrt(np.arange(days_forward) + 1)

        normal_cone = pd.DataFrame(columns=pd.Float64Index([]))
        for s in cone_stdevs:
            normal_cone[s] = midline + s * stdev
            normal_cone[-s] = midline - s * stdev

        bootstrap_cone = timeseries.forecast_cone_bootstrap(
            rets,
            days_forward,
            cone_stdevs,
            starting_value=1,
            random_seed=random_seed,
            num_samples=10000,
        )

        for col, vals in bootstrap_cone.iteritems():
            expected = normal_cone[col].values
            assert_allclose(vals.values, expected, rtol=0.005)


class TestBootstrap(TestCase):
    @parameterized.expand(
        [
            (0.0, 1.0, 1000),
            (1.0, 2.0, 500),
            (-1.0, 0.1, 10),
        ]
    )
    def test_calc_bootstrap(self, true_mean, true_sd, n):
        """Compare bootstrap distribution of the mean to sampling distribution
        of the mean.

        """
        np.random.seed(123)
        func = np.mean
        returns = pd.Series((np.random.randn(n) * true_sd) + true_mean)

        samples = timeseries.calc_bootstrap(func, returns, n_samples=10000)

        # Calculate statistics of sampling distribution of the mean
        mean_of_mean = np.mean(returns)
        sd_of_mean = np.std(returns) / np.sqrt(n)

        assert_almost_equal(
            np.mean(samples),
            mean_of_mean,
            3,
            "Mean of bootstrap does not match theoretical mean of"
            "sampling distribution",
        )

        assert_almost_equal(
            np.std(samples),
            sd_of_mean,
            3,
            "SD of bootstrap does not match theoretical SD of"
            "sampling distribution",
        )


class TestGrossLev(TestCase):
    __location__ = os.path.realpath(
        os.path.join(os.getcwd(), os.path.dirname(__file__))
    )

    test_pos = to_utc(
        pd.read_csv(
            gzip.open(__location__ + "/test_data/test_pos.csv.gz"),
            index_col=0,
            parse_dates=True,
        )
    )
    test_gross_lev = pd.read_csv(
        gzip.open(__location__ + "/test_data/test_gross_lev.csv.gz"),
        index_col=0,
        parse_dates=True,
    )
    test_gross_lev = to_series(to_utc(test_gross_lev))

    def test_gross_lev_calculation(self):
        assert_series_equal(
            timeseries.gross_lev(self.test_pos)["2004-02-01":],
            self.test_gross_lev["2004-02-01":],
            check_names=False,
        )
